# This file is part of Mach-O Loader for CLE.
# Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/) and updated in September 2019.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from cle import AT
from cle.backends.backend import Backend
from cle.backends.symbol import Symbol, SymbolType

if TYPE_CHECKING:
    from . import MachO

log = logging.getLogger(name=__name__)

# some constants:
SYMBOL_TYPE_UNDEF = 0x0
SYMBOL_TYPE_ABS = 0x2
SYMBOL_TYPE_SECT = 0xE
SYMBOL_TYPE_PBUD = 0xC
SYMBOL_TYPE_INDIR = 0xA

LIBRARY_ORDINAL_SELF = 0x0
LIBRARY_ORDINAL_OLD_MAX = 0xFE
LIBRARY_ORDINAL_DYN_LOOKUP = 0xFE

BIND_SPECIAL_DYLIB_SELF = 0x0
BIND_SPECIAL_DYLIB_WEAK_LOOKUP = 0xFD
BIND_SPECIAL_DYLIB_FLAT_LOOKUP = 0xFE
BIND_SPECIAL_DYLIB_MAIN_EXECUTABLE = 0xFF  # technically -1


class AbstractMachOSymbol(Symbol):
    """
    Base class for Mach-O symbols.
    Defines the minimum common properties all types of mach-o symbols must have
    """

    owner: MachO

    def __init__(self, owner: Backend, name: str, relative_addr: int, size: int, sym_type: SymbolType):
        super().__init__(owner, name, relative_addr, size, sym_type)

        # additional properties
        self.bind_xrefs = []  # XREFs discovered during binding of the symbol
        # starting addresses of stubs that resolve to this symbol - note that this
        # must be obtained through an analysis of some sort
        self.symbol_stubs = []

    @property
    def library_ordinal(self):
        return None

    @property
    def is_stab(self):
        return False

    @property
    def library_name(self) -> bytes | None:
        return None

    @property
    def library_base_name(self) -> str | None:
        full_name = self.library_name
        if full_name is None:
            return None

        return full_name.decode().rsplit("/", 1)[-1]


class SymbolTableSymbol(AbstractMachOSymbol):
    """
    "Regular" symbol. Made to be (somewhat) compatible with backends.Symbol.
    A SymbolTableSymbol is an entry in the binary's symbol table.

    Note that ELF-specific fields from backends.Symbol are not used and semantics of the remaining fields differ in
    many cases. As a result most stock functionality from Angr and related libraries WILL NOT WORK PROPERLY on
    MachOSymbol.

    Much of the code below is based on heuristics as official documentation is sparse, consider yourself warned!

    The relevant struct with documentation is nlist_64 defined in mach-o/nlist.h

    """

    def __init__(self, owner: MachO, symtab_offset, n_strx, n_type, n_sect, n_desc, n_value):
        # Note 1: Setting size = owner.arch.bytes has been directly taken over from the PE backend,
        # there is no meaningful definition of a symbol's size so I assume the size of an address counts here
        # Note 2: relative_addr will be the address of a symbols __got or __nl_symbol_ptr entry, not the addr of a stub
        # pointing to the symobl.
        # Stub addresses must be obtained through some sort of higher-level analysis
        # Note 3: A symbols name may not be unique!
        # Note 4: The symbol type of all symbols is SymbolType.TYPE_OTHER
        # because without docs I was unable to proplerly map Mach-O symbol types to CLE's notion of a symbol type

        # store the mach-o properties, all these are raw values straight from the binary
        self.symtab_offset = symtab_offset  # offset from the start of the symbol table
        self.n_type = n_type  # n_type field from the symbol table
        self.n_sect = n_sect  # n_sect field from the symbol table
        self.n_desc = n_desc  # n_desc  field from the symbol table
        self.n_value = n_value  # n_value field from the symbol table.
        self.n_strx = n_strx  # index into the string table

        # The meaning of n_value isn't always an address, and the logic for this is complicated
        # For now this is mostly a heuristic
        if n_value == 0:
            # This symbol doesn't have an address
            addr = 0
        elif n_value == 0x5614542:
            # This is the radr://5614542 symbol
            # The logic around this is somewhat explained here:
            # https://opensource.apple.com/source/cctools/cctools-782/misc/strip.c
            # The addr isn't really an address, but a magic value that is used to indicate that the symbol is
            addr = 0x5614542
            # addr = 0
        else:
            # The n_value is probably an address, but we need to convert it to a relative address
            addr = AT.from_lva(n_value, owner).to_rva()

        # now we may call super
        # however we cannot access any properties yet that would touch superclass-initialized attributes
        # so we have to repeat some work
        super().__init__(
            owner,
            owner.get_string(n_strx).decode("utf-8") if n_strx != 0 else "",
            addr,
            owner.arch.bytes,
            SymbolType.TYPE_OTHER,
        )

        # set further fields
        self.is_import = (
            self.sym_type == SYMBOL_TYPE_UNDEF and self.is_external and self.library_ordinal != LIBRARY_ORDINAL_SELF
        )
        self.is_export = self.name in self.owner.exports_by_name

    @property
    def library_name(self) -> bytes | None:
        if self.is_import:
            if LIBRARY_ORDINAL_DYN_LOOKUP == self.library_ordinal:
                log.warning("LIBRARY_ORDINAL_DYN_LOOKUP found, cannot handle")
                return None
            else:
                return self.owner.imported_libraries[self.library_ordinal]
        return None

    @property
    def segment_name(self):
        if self.sym_type == SYMBOL_TYPE_SECT:
            return self.owner.sections_by_ordinal[self.n_sect].segname
        else:
            return None

    @property
    def section_name(self):
        if self.sym_type == SYMBOL_TYPE_SECT:
            return self.owner.sections_by_ordinal[self.n_sect].sectname
        else:
            return None

    @property
    def value(self):
        if self.sym_type == SYMBOL_TYPE_INDIR:
            return 0
        else:
            return self.n_value

    @property
    def referenced_symbol_index(self):
        """For indirect symbols n_value contains an index into the string table indicating the referenced
        symbol's name"""
        if self.sym_type == SYMBOL_TYPE_INDIR:
            return self.n_value
        else:
            return None

    def is_weak(self):
        # compare https://developer.apple.com/library/mac/documentation/DeveloperTools/Conceptual/MachOTopics/1-Articles/executing_files.html
        return self.is_weak_referenced

    @property
    def is_function(self):
        # Incompatibility to CLE
        log.debug("It is not possible to decide wether a symbol is a function or not for MachOSymbols")
        return False

    # real symbols have properties, mach-o symbols have plenty of them:
    @property
    def is_stab(self):
        return self.n_type & 0xE0

    @property
    def is_private_external(self):
        return self.n_type & 0x10

    @property
    def is_external(self):
        return self.n_type & 0x01

    @property
    def sym_type(self):  # cannot be called "type" as that shadows a builtin variable from Symbol
        if self.is_stab:
            return self.n_type
        else:
            return self.n_type & 0x0E

    @property
    def is_common(self):
        return self.is_external and self.sym_type == SYMBOL_TYPE_UNDEF and self.n_value != 0

    @property
    def common_align(self):
        return None if not self.is_common else ((self.n_desc) >> 8) & 0x0F

    @property
    def reference_type(self):
        return self.n_desc & 0x7 if self.sym_type == SYMBOL_TYPE_UNDEF else None

    @property
    def library_ordinal(self):
        return ((self.n_desc) >> 8) & 0xFF

    @property
    def is_no_dead_strip(self):
        return self.n_desc & 0x0020

    @property
    def is_desc_discarded(self):
        return self.is_no_dead_strip

    @property
    def is_weak_referenced(self):
        return self.n_desc & 0x0040

    @property
    def is_weak_defined(self):
        return self.n_desc & 0x0080

    @property
    def is_reference_to_weak(self):
        return self.n_desc & 0x0080

    @property
    def is_thumb_definition(self):
        return self.n_desc & 0x0008

    @property
    def is_symbol_resolver(self):
        return self.n_desc & 0x0100

    @property
    def is_alt_entry(self):
        return self.n_desc & 0x0200


class DyldBoundSymbol(AbstractMachOSymbol):
    """
    The new kind of symbol handling introduced with ios15
    """

    owner: MachO

    def __init__(self, owner, name, lib_ordinal):
        """Based on the constructor of BindingSymbol"""

        # store the mach-o properties, all these are raw values straight from the binary
        self.lib_ordinal = lib_ordinal

        super().__init__(owner, name, 0, owner.arch.bytes, SymbolType.TYPE_OTHER)

        # set further fields
        self.is_import = True  # TODO: this is always an import for now
        # with ios15 came a new load command LC_DYLD_EXPORTS_TRIE
        # this isn't handled yet, so for now no symbol generated by the new dyld handling is an export
        self.is_export = False

    @property
    def library_name(self):
        if BIND_SPECIAL_DYLIB_FLAT_LOOKUP == self.lib_ordinal:
            log.warning("BIND_SPECIAL_DYLIB_FLAT_LOOKUP found, cannot handle")
            return None
        elif BIND_SPECIAL_DYLIB_WEAK_LOOKUP == self.lib_ordinal:
            return None
        try:
            return self.owner.imported_libraries[self.lib_ordinal]
        except IndexError:
            log.error(
                "Symbol %s has library ordinal %d, but there are only %d imported libraries",
                self,
                self.lib_ordinal,
                len(self.owner.imported_libraries),
            )
            return None

    @property
    def is_function(self):
        # Incompatibility to CLE
        log.debug("It is not possible to decide wether a symbol is a function or not for MachOSymbols")
        return False

    def demangled_name(self):
        return self.name  # it is not THAT easy with Mach-O

    @property
    def library_ordinal(self):
        return self.lib_ordinal


class BindingSymbol(AbstractMachOSymbol):
    """
    "Binding" symbol. Made to be (somewhat) compatible with backends.Symbol.
    A BindingSymbol is an imported symbol discovered during the binding process.

    Note that ELF-specific fields from backends.Symbol are not used and semantics of the remaining fields differ in
    many cases. As a result most stock functionality from Angr and related libraries WILL NOT WORK PROPERLY on
    MachOSymbol.

    Much of the code below is based on heuristics as official documentation is sparse, consider yourself warned!
    """

    def __init__(self, owner, name, lib_ordinal):
        # Note 1: Setting size = owner.arch.bytes has been directly taken over from the PE backend,
        # there is no meaningful definition of a symbol's size so I assume the size of an address counts here
        # Note 2: relative_addr will be the address of a symbols __got or __nl_symbol_ptr entry, not the addr of a stub
        # pointing to the symobl.
        # Stub addresses must be obtained through some sort of higher-level analysis
        # Note 3: A symbols name may not be unique!
        # Note 4: The symbol type of all symbols is SymbolType.TYPE_OTHER because without docs I was unable to problerly
        # map Mach-O symbol types to CLE's notion of a symbol type

        # store the mach-o properties, all these are raw values straight from the binary
        self.lib_ordinal = lib_ordinal

        # now we may call super
        # however we cannot access any properties yet that would touch superclass-initialized attributes
        # so we have to repeat some work
        super().__init__(owner, name, 0, owner.arch.bytes, SymbolType.TYPE_OTHER)

        # set further fields
        self.is_import = True  # this is always an import
        self.is_export = self.name in self.owner.exports_by_name

    @property
    def library_name(self):
        if LIBRARY_ORDINAL_DYN_LOOKUP == self.lib_ordinal:
            log.warning("LIBRARY_ORDINAL_DYN_LOOKUP found, cannot handle")
            return None

        return self.owner.imported_libraries[self.lib_ordinal]

    @property
    def is_function(self):
        # Incompatibility to CLE
        log.debug("It is not possible to decide wether a symbol is a function or not for MachOSymbols")
        return False

    def demangled_name(self):
        return self.name  # it is not THAT easy with Mach-O

    @property
    def library_ordinal(self):
        return self.lib_ordinal



from __future__ import annotations

from cle.backends.backend import register_backend
from cle.backends.region import Segment

from .cgc import CGC


class FakeSegment(Segment):
    def __init__(self, start, size):
        super().__init__(0, start, 0, size)
        self.is_readable = True
        self.is_writable = True
        self.is_executable = False


class BackedCGC(CGC):
    """
    This is a backend for CGC executables that allows user provide a memory backer and a register backer as the
    initial state of the running binary.
    """

    is_default = True  # Tell CLE to automatically consider using the BackedCGC backend

    def __init__(
        self,
        *args,
        memory_backer=None,
        register_backer=None,
        writes_backer=None,
        permissions_map=None,
        current_allocation_base=None,
        **kwargs,
    ):
        """
        :param path:                    File path to CGC executable.
        :param memory_backer:           A dict of memory content, with beginning address of each segment as key and
                                        actual memory content as data.
        :param register_backer:         A dict of all register contents. EIP will be used as the entry point of this
                                        executable.
        :param permissions_map:         A dict of memory region to permission flags
        :param current_allocation_base: An integer representing the current address of the top of the CGC heap.
        """
        super().__init__(*args, **kwargs)

        self.memory_backer = memory_backer
        self.register_backer = register_backer
        self.writes_backer = writes_backer
        self.permissions_map = permissions_map
        self.current_allocation_base = current_allocation_base

        for seg in self.segments:
            if seg.is_executable:
                exec_seg_addr = seg.vaddr
                break
        else:
            raise ValueError("Couldn't find executable segment?")

        for start, _ in self.memory._backers:
            if start != exec_seg_addr:
                self.memory.remove_backer(start)

        for start, data in sorted(self.memory_backer.items()):
            existing_seg = self.find_segment_containing(start)
            if existing_seg is None:  # this is the text or data segment
                new_seg = FakeSegment(start, len(data))
                self.segments.append(new_seg)

            if start == exec_seg_addr:
                continue

            if start in self.memory:
                raise ValueError("IF THIS GETS THROWN I'M GONNA JUMP OUT THE WINDOW")

            self.memory.add_backer(start, data)

        if self.register_backer is not None and "eip" in self.register_backer:
            self._entry = self.register_backer["eip"]

    @staticmethod
    def is_compatible(stream):
        return False  # Don't use this for anything unless it's manual

    @property
    def threads(self):
        return [0]

    def thread_registers(self, thread=None):
        return self.register_backer.items()


register_backend("backedcgc", BackedCGC)



from __future__ import annotations

import logging
import time

from archinfo.arch_soot import ArchSoot, SootAddressDescriptor, SootMethodDescriptor

from cle.backends.backend import Backend
from cle.errors import CLEError

try:
    import pysoot
    from pysoot.lifter import Lifter
except ImportError:
    pysoot = None
    Lifter = None

log = logging.getLogger(name=__name__)


class Soot(Backend):
    """
    The basis backend for lifting and loading bytecode from JARs and APKs to Soot IR.

    Note that self.min_addr will be 0 and self.max_addr will be 1. Hopefully no other object will be mapped at address
    0.
    """

    def __init__(
        self,
        *args,
        entry_point=None,
        entry_point_params=(),
        input_format=None,
        additional_jars=None,
        additional_jar_roots=None,
        jni_libs_ld_path=None,
        jni_libs=None,
        android_sdk=None,
        **kwargs,
    ):
        if not pysoot:
            raise ImportError("Cannot import PySoot. The Soot backend requires PySoot.")

        if kwargs.get("has_memory", False):
            raise CLEError('The parameter "has_memory" must be False for Soot backend.')

        super().__init__(*args, has_memory=False, **kwargs)
        if self.binary is None:
            raise ValueError("Cannot use the Soot backend loading from a stream")

        # load the classes
        log.debug("Lifting to Soot IR ...")
        start_time = time.time()
        pysoot_lifter = Lifter(
            self.binary,
            input_format=input_format,
            android_sdk=android_sdk,
            additional_jars=additional_jars,
            additional_jar_roots=additional_jar_roots,
        )
        end_time = time.time()
        log.debug("Lifting completed in %ds", round(end_time - start_time, 2))
        self._classes = pysoot_lifter.classes

        # find entry method
        if entry_point:
            try:
                ep_method = self.get_soot_method(entry_point, params=entry_point_params)
                ep_method_descriptor = SootMethodDescriptor.from_soot_method(ep_method)
                self._entry = SootAddressDescriptor(ep_method_descriptor, 0, 0)
                log.debug("Entry point set to %s", self._entry)
            except CLEError:
                log.warning("Couldn't find entry point %s.", entry_point)
                self._entry = None

        self.os = "javavm"
        self.rebase_addr = None
        self.set_arch(ArchSoot())

        if jni_libs:
            # native libraries are getting loaded by adding them as a dependency of this object
            self.deps += [jni_libs] if type(jni_libs) in (str, bytes) else jni_libs
            # if available, add additional load path(s)
            if jni_libs_ld_path:
                path_list = [jni_libs_ld_path] if type(jni_libs_ld_path) in (str, bytes) else jni_libs_ld_path
                self.extra_load_path += path_list
            self.jni_support = True
        else:
            self.jni_support = False

    @property
    def max_addr(self):
        # FIXME: This is a hack to satisfy checks elsewhere that max_addr must be greater than min_addr
        return self.min_addr + 1

    @property
    def entry(self):
        return self._entry

    @property
    def classes(self):
        return self._classes

    def get_soot_class(self, cls_name, none_if_missing=False):
        """
        Get Soot class object.

        :param str cls_name: Name of the class.
        :return:             The class object.
        :rtype:              pysoot.soot.SootClass
        """
        try:
            return self._classes[cls_name]
        except KeyError:
            if none_if_missing:
                return None
            else:
                raise CLEError(f'Class "{cls_name}" does not exist.')

    def get_soot_method(self, thing, class_name=None, params=(), none_if_missing=False):
        """
        Get Soot method object.

        :param thing:           Descriptor or the method, or name of the method.
        :param str class_name:  Name of the class. If not specified, class name can be parsed from method_name.
        :return:                Soot method that satisfy the criteria.
        """

        # Step 1: Parse input
        if isinstance(thing, SootMethodDescriptor):
            method_description = {
                "class_name": thing.class_name,
                "name": thing.name,
                "params": thing.params,
            }

        elif isinstance(thing, str | bytes):
            method_name = thing

            # if class_name is not set, parse it from the method name
            if class_name is None:
                last_dot = method_name.rfind(".")
                if last_dot >= 0:
                    class_name = method_name[:last_dot]
                    method_name = method_name[last_dot + 1 :]
                else:
                    raise ValueError(f"Cannot parse class name from method {method_name}.")

            method_description = {
                "class_name": class_name,
                "name": method_name,
                "params": params,
            }

        else:
            raise TypeError(f'Unsupported type "{thing}" for the first argument.')

        # Step 2: Load class containing the method
        try:
            cls = self.get_soot_class(method_description["class_name"])
        except CLEError:
            if none_if_missing:
                return None
            else:
                raise

        # Step 3: Get all methods matching the description
        methods = [
            soot_method
            for soot_method in cls.methods
            if self._description_matches_soot_method(soot_method, **method_description)
        ]

        if not methods:
            if none_if_missing:
                return None
            else:
                raise CLEError(
                    "Method with description {} does not exist in class {}.".format(
                        method_description, method_description["class_name"]
                    )
                )

        if len(methods) > 1:
            # Warn if we found several matching methods
            log.warning(
                "Method with description %s is ambiguous in class %s.",
                method_description,
                method_description["class_name"],
            )

        return methods[0]

    @staticmethod
    def _description_matches_soot_method(soot_method, name=None, class_name=None, params=()):
        if name and soot_method.name != name:
            return False
        if class_name and soot_method.class_name != class_name:
            return False
        if soot_method.params != params:
            return False
        return True

    @property
    def main_methods(self):
        """
        Find all Main methods in this binary.

        :return: All main methods in each class.
        :rtype:  iterator
        """
        for cls in self.classes.values():
            for method in cls.methods:
                if method.name == "main":  # TODO: Check more attributes
                    yield method

    @staticmethod
    def is_zip_archive(stream):
        stream.seek(0)
        identstring = stream.read(4)
        stream.seek(0)
        return identstring.startswith(b"\x50\x4b\x03\x04")



from __future__ import annotations

import logging

from cle.address_translator import AT
from cle.backends.symbol import SymbolType
from cle.errors import CLEInvalidBinaryError, CLEOperationError

from .elfreloc import ELFReloc

log = logging.getLogger(name=__name__)


__all__ = [
    "GenericTLSDoffsetReloc",
    "GenericTLSOffsetReloc",
    "GenericTLSDescriptorReloc",
    "GenericTLSModIdReloc",
    "GenericIRelativeReloc",
    "GenericAbsoluteAddendReloc",
    "GenericPCRelativeAddendReloc",
    "GenericJumpslotReloc",
    "GenericRelativeReloc",
    "GenericAbsoluteReloc",
    "GenericCopyReloc",
    "MipsGlobalReloc",
    "MipsLocalReloc",
    "RelocTruncate32Mixin",
    "RelocGOTMixin",
]


class GenericTLSDoffsetReloc(ELFReloc):
    @property
    def value(self):
        return self.addend + self.symbol.relative_addr

    def resolve_symbol(self, solist, **kwargs):  # pylint: disable=unused-argument
        self.resolve(None)
        return True


class GenericTLSOffsetReloc(ELFReloc):
    AUTO_HANDLE_NONE = True

    def relocate(self):
        hell_offset = self.owner.arch.elf_tls.tp_offset

        if self.resolvedby is None:
            obj = self.owner
            addr = 0
        else:
            obj = self.resolvedby.owner
            addr = self.resolvedby.relative_addr

        if obj.tls_block_offset is None:
            raise CLEInvalidBinaryError("Illegal relocation - dynamically loaded object using static TLS")

        self.owner.memory.pack_word(self.relative_addr, obj.tls_block_offset + self.addend + addr - hell_offset)


class GenericTLSDescriptorReloc(ELFReloc):
    # Going VERY far out on a limb here
    # "TLS descriptors" are a thing I'm seeing in aarch64 binaries which seem to want to relocate by
    # sticking a pointer to a resolver function followed by some arbitrary data. The resolver function
    # is passed a pointer to the descriptor. My guess is the resolver is supposed to basically perform
    # _tls_get_addr, but the intention is probably to make it possible to work with dynamically loaded objects.

    RESOLVER_ADDR: int = NotImplemented
    AUTO_HANDLE_NONE = True

    def relocate(self):
        if self.resolvedby is None:
            obj = self.owner
        else:
            obj = self.resolvedby.owner

        if obj.tls_block_offset is None:
            raise CLEInvalidBinaryError("Illegal relocation? - dynamically loaded object using static TLS? Maybe?")

        self.owner.memory.pack_word(self.relative_addr, self.RESOLVER_ADDR)
        self.owner.memory.pack_word(
            self.relative_addr + self.arch.bytes, obj.tls_block_offset + self.addend + self.symbol.relative_addr
        )  # Should this include the hell offset?


class GenericTLSModIdReloc(ELFReloc):
    AUTO_HANDLE_NONE = True

    def relocate(self):
        if self.symbol.type == SymbolType.TYPE_NONE:
            obj = self.owner
        else:
            obj = self.resolvedby.owner

        self.owner.memory.pack_word(self.relative_addr, obj.tls_module_id)


class GenericIRelativeReloc(ELFReloc):
    AUTO_HANDLE_NONE = True

    def relocate(self):
        if self.symbol.type == SymbolType.TYPE_NONE:
            self.owner.irelatives.append((AT.from_lva(self.addend, self.owner).to_mva(), self.relative_addr))
        else:
            self.owner.irelatives.append((self.resolvedby.rebased_addr, self.relative_addr))


class GenericAbsoluteAddendReloc(ELFReloc):
    @property
    def value(self):
        return self.resolvedby.rebased_addr + self.addend


class GenericPCRelativeAddendReloc(ELFReloc):
    @property
    def value(self):
        return self.resolvedby.rebased_addr + self.addend - self.rebased_addr


class GenericJumpslotReloc(ELFReloc):
    @property
    def value(self):
        if self.is_rela:
            return self.resolvedby.rebased_addr + self.addend
        else:
            return self.resolvedby.rebased_addr


class GenericRelativeReloc(ELFReloc):
    AUTO_HANDLE_NONE = True

    @property
    def value(self):
        if self.resolvedby is not None:
            return self.resolvedby.rebased_addr
        return self.owner.mapped_base + self.addend


class GenericAbsoluteReloc(ELFReloc):
    @property
    def value(self):
        return self.resolvedby.rebased_addr


class GenericCopyReloc(ELFReloc):
    def resolve_symbol(self, solist, **kwargs):
        new_solist = [x for x in solist if x is not self.owner]
        super().resolve_symbol(new_solist, **kwargs)

    def relocate(self):
        if self.resolvedby.size != self.symbol.size and (self.resolvedby.size != 0 or not self.resolvedby.is_extern):
            log.error("Export symbol is different size than import symbol for copy relocation: %s", self.symbol.name)
        else:
            self.owner.memory.store(
                self.relative_addr,
                self.resolvedby.owner.memory.load(self.resolvedby.relative_addr, self.resolvedby.size),
            )
        return True


class MipsGlobalReloc(GenericAbsoluteReloc):
    pass


class MipsLocalReloc(ELFReloc):
    AUTO_HANDLE_NONE = True

    def resolve_symbol(self, solist, **kwargs):
        self.resolve(None)

    def relocate(self):
        if self.owner.mapped_base == 0:
            return  # don't touch local relocations on the main bin

        delta = self.owner.mapped_base - self.owner._dynamic["DT_MIPS_BASE_ADDRESS"]
        if delta == 0:
            return

        val = self.owner.memory.unpack_word(self.relative_addr)
        newval = val + delta
        self.owner.memory.pack_word(self.relative_addr, newval)


class RelocTruncate32Mixin:
    """
    A mix-in class for relocations that cover a 32-bit field regardless of the architecture's address word length.
    """

    # If True, 32-bit truncated value must equal to its original when zero-extended
    check_zero_extend = False

    # If True, 32-bit truncated value must equal to its original when sign-extended
    check_sign_extend = False

    def relocate(self):
        arch_bits = self.owner.arch.bits
        assert arch_bits >= 32  # 16-bit makes no sense here

        val = self.value % (2**arch_bits)  # we must truncate it to native range first

        if (
            self.check_zero_extend
            and val >> 32 != 0
            or self.check_sign_extend
            and val >> 32 != ((1 << (arch_bits - 32)) - 1)
            if ((val >> 31) & 1) == 1
            else 0
        ):
            raise CLEOperationError(
                f"relocation truncated to fit: {self.__class__.__name__}; consider making"
                " relevant addresses fit in the 32-bit address space."
            )

        self.owner.memory.pack_word(self.dest_addr, val, size=4, signed=False)

        return True


class RelocGOTMixin:
    """
    A mix-in class which will cause the symbol to be resolved to a pointer to the symbol instead of the symbol
    """

    def resolve(self, symbol, extern_object=None):
        assert extern_object is not None, "I have no idea how this would happen"

        got_symbol = extern_object.make_extern(f"got.{symbol.name}", sym_type=SymbolType.TYPE_OBJECT, point_to=symbol)
        super().resolve(got_symbol)



from __future__ import annotations

import struct
from collections import namedtuple

import archinfo

from . import Backend, Section, register_backend

HEADER = struct.Struct("<HHBBHIIQIIII")
SECTION_HEADER = struct.Struct("<8sIIIIIIHHI")
HeaderType = namedtuple(
    "HeaderType",
    (
        "signature",
        "machine",
        "number_of_sections",
        "subsystem",
        "stripped_size",
        "address_of_entry_point",
        "base_of_code",
        "image_base",
        "data_directory_0_virtual_address",
        "data_directory_0_size",
        "data_directory_1_virtual_address",
        "data_directory_1_size",
    ),
)
SectionHeaderType = namedtuple(
    "SectionHeaderType",
    (
        "section_name",
        "physical_address_virtual_size",
        "virtual_address",
        "size_of_raw_data",
        "pointer_to_raw_data",
        "pointer_to_relocations",
        "pointer_to_line_numbers",
        "number_of_relocations",
        "number_of_line_numbers",
        "characteristics",
    ),
)

ARCH_MAPPING = {
    0x014C: "i386",
    0x0200: "IPF",
    0x0EBC: "EBC",
    0x8664: "X64",
    0x01C0: "ARM Cortex-M",
    0x01C2: "ARM",
}


class TE(Backend):
    """
    A "Terse Executable" format image, commonly used as part of UEFI firmware drivers.
    """

    is_default = True

    @classmethod
    def is_compatible(cls, stream):
        stream.seek(0)
        return stream.read(2) == b"VZ"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._binary_stream.seek(0)
        self.header = HeaderType(*HEADER.unpack(self._binary_stream.read(HEADER.size)))
        self.section_headers = [
            SectionHeaderType(*SECTION_HEADER.unpack(self._binary_stream.read(SECTION_HEADER.size)))
            for _ in range(self.header.number_of_sections)
        ]

        self.set_arch(archinfo.arch_from_id(ARCH_MAPPING[self.header.machine]))

        offset_offset = self.header.stripped_size - HEADER.size
        self.linked_base = self.mapped_base = self.header.image_base + offset_offset

        has_relocs = False
        for section_header in self.section_headers:
            region = Section(
                section_header.section_name.rstrip(b"\0").decode(),
                section_header.pointer_to_raw_data,
                section_header.virtual_address + self.linked_base,
                section_header.physical_address_virtual_size,
            )
            self._sections.append(region)

            if section_header.characteristics & 0x02000000 != 0 or section_header.physical_address_virtual_size == 0:
                # discard or no data
                continue
            self._binary_stream.seek(section_header.pointer_to_raw_data - offset_offset)
            data = self._binary_stream.read(section_header.size_of_raw_data)
            assert len(data) == section_header.size_of_raw_data
            if section_header.size_of_raw_data < section_header.physical_address_virtual_size:
                data = data.ljust(section_header.physical_address_virtual_size - section_header.size_of_raw_data, b"\0")
            self.memory.add_backer(section_header.virtual_address, data)

            if section_header.number_of_relocations != 0:
                has_relocs = True

        self._segments = self._sections

        self.pic = has_relocs


register_backend("te", TE)



from __future__ import annotations

import logging
import os
import re
import struct

import archinfo
import pefile

try:
    import pyxdia
except ImportError:
    pyxdia = None

from cle.address_translator import AT
from cle.backends.backend import Backend, FunctionHint, FunctionHintSource, register_backend
from cle.backends.symbol import SymbolType
from cle.utils import extract_null_terminated_bytestr

from .regions import PESection
from .relocation import get_relocation
from .relocation.generic import IMAGE_REL_BASED_ABSOLUTE, IMAGE_REL_BASED_HIGHADJ, DllImport
from .symbol import WinSymbol

PDB_SUPPORT_ENABLED = pyxdia is not None
SECTION_NAME_STRING_TABLE_OFFSET_RE = re.compile(r"\/(\d+)")
VALID_SYMBOL_NAME_RE = re.compile(r"[A-Za-z0-9_@$?]+")

log = logging.getLogger(name=__name__)


class PE(Backend):
    """
    Representation of a PE (i.e. Windows) binary.

    Useful backend options:

    - ``debug_symbols``: Provides the path to a PDB file which contains the binary's debug symbols
    """

    is_default = True  # Tell CLE to automatically consider using the PE backend

    def __init__(self, *args, debug_symbols=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_load_args(debug_symbols=debug_symbols)
        self.segments = self.sections  # in a PE, sections and segments have the same meaning
        self.os = "windows"
        self._raw_data = self._binary_stream.read()
        if self.binary is None:
            self._pe = pefile.PE(data=self._raw_data, fast_load=True)
            self._parse_pe_non_reloc_data_directories()
        elif self.binary in self._pefile_cache:  # these objects are not mutated, so they are reusable within a process
            self._pe = self._pefile_cache[self.binary]
        else:
            self._pe = pefile.PE(self.binary, fast_load=True)
            self._parse_pe_non_reloc_data_directories()
            if not self.is_main_bin:
                # only cache shared libraries, the main binary will not be reused
                self._pefile_cache[self.binary] = self._pe

        assert self._pe.FILE_HEADER is not None
        assert self._pe.OPTIONAL_HEADER is not None

        if self._arch is None:
            machine_type = self._pe.FILE_HEADER.Machine
            self.set_arch(archinfo.arch_from_id(pefile.MACHINE_TYPE.get(machine_type, hex(machine_type))))

        self.mapped_base = self.linked_base = self._pe.OPTIONAL_HEADER.ImageBase

        self._entry = AT.from_rva(self._pe.OPTIONAL_HEADER.AddressOfEntryPoint, self).to_lva()

        if hasattr(self._pe, "DIRECTORY_ENTRY_IMPORT"):
            self.deps = [entry.dll.decode().lower() for entry in self._pe.DIRECTORY_ENTRY_IMPORT]
        else:
            self.deps = []

        if self.binary is not None and not self.is_main_bin:
            self.provides = os.path.basename(self.binary).lower()
        else:
            self.provides = None

        self.tls_index_address = None
        self.tls_callbacks = None

        self.supports_nx = self._pe.OPTIONAL_HEADER.DllCharacteristics & 0x100 != 0
        self.pic = self.pic or self._pe.OPTIONAL_HEADER.DllCharacteristics & 0x40 != 0
        if hasattr(self._pe, "DIRECTORY_ENTRY_LOAD_CONFIG"):
            self.load_config = {
                name: value["Value"]
                for name, value in self._pe.DIRECTORY_ENTRY_LOAD_CONFIG.struct.dump_dict().items()
                if name != "Structure"
            }
        else:
            self.load_config = {}

        self._exports = {}
        self._ordinal_exports = {}
        self._symbol_cache = self._exports  # same thing
        self._handle_imports()
        self._handle_exports()
        self._handle_seh()
        if self.loader._perform_relocations:
            # parse base relocs
            self._pe.parse_data_directories(directories=[pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_BASERELOC"]])
            self.__register_relocs()
        # parse TLS
        self._register_tls()
        # parse sections
        self._register_sections()

        self.linking = "dynamic" if self.deps else "static"
        self.jmprel = self._get_jmprel()
        self.memory.add_backer(0, self._pe.get_memory_mapped_image(max_virtual_address=0x100000000))

        if debug_symbols or self.loader._load_debug_info:
            pdb_path = debug_symbols or self._find_pdb_path()
            if pdb_path:
                self.load_symbols_from_pdb(pdb_path)

        self._load_symbols_from_coff_header()

        self.is_dotnet = (
            self._pe.OPTIONAL_HEADER.DATA_DIRECTORY[
                pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR"]
            ].VirtualAddress
            != 0
        )

    _pefile_cache = {}

    @classmethod
    def is_compatible(cls, stream):
        identstring = stream.read(0x1000)
        stream.seek(0)
        if identstring.startswith(b"MZ") and len(identstring) > 0x40:
            peptr = struct.unpack("I", identstring[0x3C:0x40])[0]
            if peptr < len(identstring) and identstring[peptr : peptr + 4] == b"PE\0\0":
                return True
        return False

    @classmethod
    def check_magic_compatibility(cls, stream):
        stream.seek(0)
        identstring = stream.read(0x10)
        stream.seek(0)
        return identstring.startswith(b"MZ")

    @classmethod
    def check_compatibility(cls, spec, obj):
        if hasattr(spec, "read") and hasattr(spec, "seek"):
            pe = pefile.PE(data=spec.read(), fast_load=True)
        else:
            pe = pefile.PE(spec, fast_load=True)

        assert pe.FILE_HEADER is not None

        arch = archinfo.arch_from_id(pefile.MACHINE_TYPE[pe.FILE_HEADER.Machine])
        return arch == obj.arch

    #
    # Public methods
    #

    def close(self):
        super().close()
        del self._pe
        del self._raw_data

    def get_symbol(self, name):
        """
        Look up the symbol with the given name. Symbols can be looked up by ordinal with the name ``"ordinal.%d" % num``
        """
        if name.startswith("ordinal."):
            return self._ordinal_exports.get(int(name.split(".")[1]), None)
        return super().get_symbol(name)

    def load_symbols_from_pdb(self, pdb_path):
        """
        Load available symbols from PDB at `pdb_path`
        """
        if pyxdia is None:
            log.warning("Install pyxdia to load symbols from %s", pdb_path)
            return

        log.debug("Loading symbols from %s", pdb_path)
        try:
            pdb = pyxdia.PDB(pdb_path)
        except:  # noqa:E722 pylint:disable=bare-except
            log.exception("Failed to load PDB at %s", pdb_path)
            return

        for global_ in pdb.globals:
            rva = global_["relativeVirtualAddress"]
            if rva is None:
                continue
            name = global_["name"]
            tag = str(global_["symTag"])
            symbol_type = {
                "Data": SymbolType.TYPE_OBJECT,
                "Function": SymbolType.TYPE_FUNCTION,
            }.get(tag, SymbolType.TYPE_OTHER)
            symb = WinSymbol(self, name, rva, False, False, None, None, symbol_type)
            log.debug("Adding symbol %s", str(symb))
            self.symbols.add(symb)

    #
    # Private methods
    #

    def _parse_pe_non_reloc_data_directories(self):
        """
        Parse data directories that is not DIRECTORY_ENTRY_BASERELOC since parsing relocations can take a long time in
        many PE binaries.
        """

        directory_names = (
            "IMAGE_DIRECTORY_ENTRY_EXPORT",
            "IMAGE_DIRECTORY_ENTRY_IMPORT",
            "IMAGE_DIRECTORY_ENTRY_RESOURCE",
            "IMAGE_DIRECTORY_ENTRY_EXCEPTION",
            "IMAGE_DIRECTORY_ENTRY_SECURITY",
            "IMAGE_DIRECTORY_ENTRY_DEBUG",
            "IMAGE_DIRECTORY_ENTRY_COPYRIGHT",
            "IMAGE_DIRECTORY_ENTRY_GLOBALPTR",
            "IMAGE_DIRECTORY_ENTRY_TLS",
            "IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG",
            "IMAGE_DIRECTORY_ENTRY_IAT",
            "IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT",
            "IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR",
            "IMAGE_DIRECTORY_ENTRY_RESERVED",
        )
        directories = [pefile.DIRECTORY_ENTRY[n] for n in directory_names]
        self._pe.parse_data_directories(directories=directories)

    def _get_jmprel(self):
        return self.imports

    def _handle_imports(self):
        if hasattr(self._pe, "DIRECTORY_ENTRY_IMPORT"):
            for entry in self._pe.DIRECTORY_ENTRY_IMPORT:
                for imp in entry.imports:
                    if imp.name is None:  # must be an import by ordinal
                        imp_name = f"ordinal.{imp.ordinal}.{entry.dll.lower().decode()}"
                    else:
                        imp_name = imp.name.decode()

                    symb = WinSymbol(
                        owner=self,
                        name=imp_name,
                        addr=0,
                        is_import=True,
                        is_export=False,
                        ordinal_number=imp.ordinal,
                        forwarder=None,
                    )
                    self.symbols.add(symb)
                    reloc = self._make_reloc(
                        addr=AT.from_lva(imp.address, self).to_rva(),
                        reloc_type=None,
                        symbol=symb,
                        resolvewith=entry.dll.decode(),
                    )

                    if reloc is not None:
                        self.imports[imp_name] = reloc
                        self.relocs.append(reloc)

    def _handle_exports(self):
        if hasattr(self._pe, "DIRECTORY_ENTRY_EXPORT"):
            symbols = self._pe.DIRECTORY_ENTRY_EXPORT.symbols
            for exp in symbols:
                name = exp.name.decode() if exp.name is not None else None
                forwarder = exp.forwarder.decode() if exp.forwarder is not None else None
                symb = WinSymbol(self, name, exp.address, False, True, exp.ordinal, forwarder)
                self.symbols.add(symb)
                self._exports[name] = symb
                self._ordinal_exports[exp.ordinal] = symb

                if forwarder is not None:
                    forwardlib = forwarder.split(".", 1)[0].lower() + ".dll"
                    if forwardlib not in self.deps:
                        self.deps.append(forwardlib)

    def _handle_seh(self):
        if hasattr(self._pe, "DIRECTORY_ENTRY_EXCEPTION"):
            for entry in self._pe.DIRECTORY_ENTRY_EXCEPTION:
                self.function_hints.append(
                    FunctionHint(
                        entry.struct.BeginAddress + self.linked_base,
                        entry.struct.EndAddress - entry.struct.BeginAddress,
                        FunctionHintSource.EH_FRAME,
                    )
                )

    def __register_relocs(self):
        if not hasattr(self._pe, "DIRECTORY_ENTRY_BASERELOC"):
            log.debug("%s has no relocations", self.binary)
            return []

        for base_reloc in self._pe.DIRECTORY_ENTRY_BASERELOC:
            entry_idx = 0
            while entry_idx < len(base_reloc.entries):
                reloc_data = base_reloc.entries[entry_idx]
                if (
                    reloc_data.type == pefile.RELOCATION_TYPE["IMAGE_REL_BASED_HIGHADJ"]
                ):  # special case, occupies 2 entries
                    if entry_idx == len(base_reloc.entries):
                        log.warning("PE contains corrupt base relocation table")
                        break

                    next_entry = base_reloc.entries[entry_idx]
                    entry_idx += 1
                    reloc = self._make_reloc(addr=reloc_data.rva, reloc_type=reloc_data.type, next_rva=next_entry.rva)
                else:
                    reloc = self._make_reloc(addr=reloc_data.rva, reloc_type=reloc_data.type)

                if reloc is not None:
                    # Some binaries have the DYNAMIC_BASE DllCharacteristic unset but have tons of fixup relocations
                    self.pic = True
                    self.relocs.append(reloc)

                entry_idx += 1

        return self.relocs

    def _make_reloc(self, addr, reloc_type, symbol=None, next_rva=None, resolvewith=None):
        # Handle special cases first
        if reloc_type == 0:  # 0 simply means "ignore this relocation"
            reloc = IMAGE_REL_BASED_ABSOLUTE(owner=self, symbol=symbol, addr=addr, resolvewith=resolvewith)
            return reloc
        if reloc_type is None:  # for DLL imports
            reloc = DllImport(owner=self, symbol=symbol, addr=addr, resolvewith=resolvewith)
            return reloc
        if next_rva is not None:
            reloc = IMAGE_REL_BASED_HIGHADJ(owner=self, addr=addr, next_rva=next_rva)
            return reloc

        # Handle all the normal base relocations
        RelocClass = get_relocation(self.arch.name, reloc_type)
        if RelocClass is None:
            log.debug("Failed to find relocation class for arch %s, type %d", "pe" + self.arch.name, reloc_type)
            return None

        cls = RelocClass(owner=self, symbol=symbol, addr=addr)
        if cls is None:
            log.warning(
                "Failed to retrieve relocation for %s of type %s",
                symbol.name if symbol else "<unknown symbol>",
                reloc_type,
            )

        return cls

    def _register_tls(self):
        if hasattr(self._pe, "DIRECTORY_ENTRY_TLS"):
            tls = self._pe.DIRECTORY_ENTRY_TLS.struct

            self.tls_used = True
            self.tls_data_start = AT.from_lva(tls.StartAddressOfRawData, self).to_rva()
            self.tls_data_size = tls.EndAddressOfRawData - tls.StartAddressOfRawData
            self.tls_index_address = tls.AddressOfIndex
            self.tls_callbacks = self._register_tls_callbacks(tls.AddressOfCallBacks)
            self.tls_block_size = self.tls_data_size + tls.SizeOfZeroFill

    def _register_tls_callbacks(self, addr):
        """
        TLS callbacks are stored as an array of virtual addresses to functions.
        The last entry is empty (NULL), which indicates the end of the table
        """
        callbacks = []

        callback_rva = AT.from_lva(addr, self).to_rva()
        callback = self._pe.get_dword_at_rva(callback_rva)
        while callback != 0 and callback is not None:
            callbacks.append(callback)
            callback_rva += 4
            callback = self._pe.get_dword_at_rva(callback_rva)

        return callbacks

    def _read_from_string_table(self, offset: int, encoding: str = "latin-1") -> str:
        """
        Read a null-terminated string from the string table given a byte offset.

        :param offset: Byte offset of the string.
        :param encoding: String encoding (default latin-1).
        """
        assert self._pe.FILE_HEADER is not None
        offset += self._pe.FILE_HEADER.PointerToSymbolTable + self._pe.FILE_HEADER.NumberOfSymbols * 18
        return extract_null_terminated_bytestr(self._raw_data, offset).decode(encoding)

    def _register_sections(self):
        """
        Wrap self._pe.sections in PESection objects, and add them to self.sections.
        """

        for pe_section in self._pe.sections:
            name = pe_section.Name.rstrip(b"\x00").decode("latin-1")
            # Match indirect section names given by a forward slash and a
            # decimal byte offset into the string table.
            str_tbl_offset_match = SECTION_NAME_STRING_TABLE_OFFSET_RE.fullmatch(name)
            if str_tbl_offset_match:
                str_tbl_offset = int(str_tbl_offset_match.group(1))
                name = self._read_from_string_table(str_tbl_offset)
            section = PESection(pe_section, remap_offset=self.linked_base, name=name)
            self.sections.append(section)
            self.sections_map[section.name] = section

    def _find_pdb_path(self):
        """
        Find path to the PDB file containing debug information for this binary.
        """
        path = None
        checks = []

        # Check PE file for path to PDB
        if hasattr(self._pe, "DIRECTORY_ENTRY_DEBUG"):
            for de in self._pe.DIRECTORY_ENTRY_DEBUG:
                if de.entry and hasattr(de.entry, "PdbFileName"):
                    path = de.entry.PdbFileName.rstrip(b"\x00").decode()
                    break

        if path:
            if os.path.exists(path):
                return path
            checks.append(path)

            # PDB not at specified location; check next to binary
            if self.binary:
                filename = os.path.basename(path.replace("\\", "/"))
                path = os.path.join(os.path.dirname(self.binary), filename)
                if os.path.exists(path):
                    return path
                checks.append(path)

        # Guess PDB has same name as binary
        if self.binary:
            path = os.path.splitext(self.binary)[0] + ".pdb"
            if os.path.exists(path):
                return path
            checks.append(path)

        log.warning("Unable to find PDB file for this PE. Tried: %s", str(checks))
        return None

    def _load_symbols_from_coff_header(self):
        """
        COFF debug info is deprecated, but may still be provided (e.g. by mingw).
        """
        type_to_symbol_type = {
            0: SymbolType.TYPE_OBJECT,
            0x20: SymbolType.TYPE_FUNCTION,
        }

        assert self._pe.FILE_HEADER is not None

        sizeof_symbol_desc = 18

        # Verify symbol table is within file bounds
        end_of_table_offset = (
            self._pe.FILE_HEADER.PointerToSymbolTable + self._pe.FILE_HEADER.NumberOfSymbols * sizeof_symbol_desc
        )
        if end_of_table_offset >= len(self._raw_data):
            log.warning("PE symbol table out of bounds")
            return

        idx = 0
        while idx < self._pe.FILE_HEADER.NumberOfSymbols:
            offset = self._pe.FILE_HEADER.PointerToSymbolTable + idx * sizeof_symbol_desc
            sym_desc = self._raw_data[offset : offset + sizeof_symbol_desc]
            (name, value, section, type_, _, num_aux_syms) = struct.unpack("<8sIhHBB", sym_desc)
            name_as_dwords = struct.unpack("<II", name)
            if name_as_dwords[0] == 0:
                name = self._read_from_string_table(name_as_dwords[1])
            else:
                name = name.rstrip(b"\x00").decode("latin-1")
            if section > 0 and type_ in type_to_symbol_type and VALID_SYMBOL_NAME_RE.fullmatch(name):
                rva = self._pe.sections[section - 1].VirtualAddress + value
                symbol = WinSymbol(self, name, rva, False, False, None, None, type_to_symbol_type[type_])
                log.debug("Adding symbol %s", symbol)
                self.symbols.add(symbol)
            idx += 1 + num_aux_syms


register_backend("pe", PE)



from __future__ import annotations

from collections import defaultdict

from cle.backends.relocation import Relocation
from cle.backends.symbol import Symbol, SymbolType


class SimData(Symbol):
    """
    A SimData class is used to provide data when there is an unresolved data import symbol.

    To use it, subclass this class and implement the below attributes and methods.

    :cvar name:     The name of the symbol to provide
    :cvar libname:  The name of the library from which the symbol originally comes (currently unused).
    :cvar type:     The type of the symbol, usually ``SymbolType.TYPE_OBJECT``.

    Use the below `register` method to register SimData subclasses with CLE.

    NOTE: SimData.type hides the Symbol.type instance property
    """

    name: str = NotImplemented
    type: SymbolType = NotImplemented
    libname: str = NotImplemented

    @classmethod
    def static_size(cls, owner) -> int:
        """
        Implement me: return the size of the symbol in bytes before it gets constructed

        :param owner:   The ExternObject owning the symbol-to-be. Useful to get at ``owner.arch``.
        """
        return NotImplemented

    def value(self) -> bytes:
        """
        Implement me: the initial value of the bytes in memory for the symbol. Should return a
        bytestring of the same length as static_size returned. (owner is ``self.owner`` now)
        """
        return NotImplemented

    def relocations(self) -> list[Relocation]:
        """
        Maybe implement me: If you like, return a list of relocation objects to apply. To create
        new import symbols, use ``self.owner.make_extern_import``.
        """
        return []


registered_data = defaultdict(list)


def register(simdata_cls: type[SimData]):
    """
    Register the given SimData class with CLE so it may be used during loading
    """
    if simdata_cls.name is None:
        return
    registered_data[simdata_cls.name].append(simdata_cls)


def lookup(name: str, libname) -> type[SimData] | None:
    weak_option = None
    for simdata_cls in registered_data[name]:
        if type(libname) is type(simdata_cls.libname) is str and simdata_cls.libname.startswith(libname):  # noqa: E721
            return simdata_cls
        elif simdata_cls is None or libname is None:
            weak_option = simdata_cls

    return weak_option



"""
Constant values for lifecycle of Apk.
"""

from __future__ import annotations

component_class = [
    "android.app.Activity",
    "android.app.Service",
    "android.content.BroadcastReceiver",
    "android.content.ContentProvider",
    "com.google.android.gcm.GCMBaseIntentService",
    "com.google.android.gms.gcm.GcmListenerService",
    "android.app.Application",
    "android.app.Fragment",
    "android.support.v4.app.Fragment",
    "androidx.fragment.app.Fragment",
    "android.content.ServiceConnection",
    "com.google.android.maps.MapActivity",
    "android.support.v4.app.AppCompatActivity",
    "android.support.v7.app.AppCompatActivity",
    "androidx.appcompat.app.AppCompatActivity",
    "android.content.ComponentCallbacks",
    "android.content.ComponentCallbacks2",
]
callback = {
    "activity": [
        "onCreate(android.os.Bundle)",
        "onStart()",
        "onRestoreInstanceState(android.os.Bundle)",
        "onPostCreate(android.os.Bundle)",
        "onResume()",
        "onPostResume()",
        "onCreateDescription()",
        "onSaveInstanceState(android.os.Bundle)",
        "onPause()",
        "onStop()",
        "onRestart()",
        "onDestroy()",
        "onAttachFragment(android.app.Fragment)",
    ],
    "service": [
        "onCreate()",
        "onStart(android.content.Intent,int)",
        "onStartCommand(android.content.Intent,int,int)",
        "onBind(android.content.Intent)",
        "onRebind(android.content.Intent)",
        "onUnbind(android.content.Intent)",
        "onDestroy()",
    ],
    "receiver": [
        "onReceive(android.content.Context,android.content.Intent)",
    ],
    "provider": [
        "onCreate()",
        "insert(android.net.Uri,android.content.ContentValues)",
        "query(android.net.Uri,java.lang.String[],java.lang.String,java.lang.String[],java.lang.String)",
        "update(android.net.Uri,android.content.ContentValues,java.lang.String,java.lang.String[])",
        "delete(android.net.Uri,java.lang.String,java.lang.String[])",
        "getType(android.net.Uri)",
    ],
    "gcmintentservice": [
        "onDeletedMessages(android.content.Context,int)",
        "onError(android.content.Context,java.lang.String)",
        "onMessage(android.content.Context,android.content.Intent)",
        "onRecoverableError(android.content.Context,java.lang.String)",
        "onRegistered(android.content.Context,java.lang.String)",
        "onUnregistered(android.content.Context,java.lang.String)",
    ],
    "gcmlistenerservice": [
        "onDeletedMessages()",
        "onMessageReceived(java.lang.String,android.os.Bundle)",
        "onMessageSent(java.lang.String)",
        "onSendError(java.lang.String,java.lang.String)",
    ],
    "application": [
        "onCreate()",
        "onTerminate()",
        "onServiceConnected(android.content.ComponentName,android.os.IBinder)",
        "onServiceDisconnected(android.content.ComponentName)",
        "android.app.Application$ActivityLifecycleCallbacks",
        "onActivityStarted(android.app.Activity)",
        "onActivityStopped(android.app.Activity)",
        "onActivitySaveInstanceState(android.app.Activity,android.os.Bundle)",
        "onActivityResumed(android.app.Activity)",
        "onActivityPaused(android.app.Activity)",
        "onActivityDestroyed(android.app.Activity)",
        "onActivityCreated(android.app.Activity,android.os.Bundle)",
        "onLowMemory()",
        "onConfigurationChanged(android.content.res.Configuration)",
        "onTrimMemory(int)",
    ],
    "fragment": [
        "onCreate(android.os.Bundle)",
        "onAttach(android.app.Activity)",
        "onCreateView(android.view.LayoutInflater,android.view.ViewGroup,android.os.Bundle)",
        "onViewCreated(android.view.View,android.os.Bundle)",
        "onStart()",
        "onActivityCreated(android.os.Bundle)",
        "onViewStateRestored(android.app.Activity)",
        "onResume()",
        "onPause()",
        "onStop()",
        "onDestroyView()",
        "onDestroy()",
        "onDetach()",
        "onSaveInstanceState(android.os.Bundle)",
    ],
}



"""Relocation types for ARM.

Reference: https://github.com/ARM-software/abi-aa/blob/main/aaelf32/aaelf32.rst#relocation-codes
"""

from __future__ import annotations

import logging

from cle.errors import CLEOperationError

from .elfreloc import ELFReloc
from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
    RelocGOTMixin,
    RelocTruncate32Mixin,
)

log = logging.getLogger(name=__name__)


def _applyReloc(inst, result, mask=0xFFFFFFFF):
    """
    Applies the specified mask to the relocation and verifies that the mask
    is valid for the given result.
    """
    try:
        if result & ~mask:
            raise ValueError("result & ~mask is not 0.")
    except ValueError as ex:
        log.warning("Relocation failed: %r", ex)
        return 0  # worst case, you hook it yourself
    return (inst & ~mask) | (result & mask)  # pylint: disable=superfluous-parens


def _isThumbFunc(symbol, addr):
    """
    Checks whether the provided symbol and address is a Thumb function by
    verifying the LSB is 1 and the symbol is STT_FUNC.
    """
    return (addr % 2 == 1) and symbol.is_function


class R_ARM_CALL(ELFReloc):
    """
    Relocate R_ARM_CALL symbols via instruction modification. It additionally
    handles R_ARM_PC24 and R_ARM_JUMP24. The former is deprecated and is now
    just the same as R_ARM_CALL.

    R_ARM_JUMP24 doesn't need the Thumb check. Technically, if the Thumb check
    succeeds on R_ARM_JUMP24, it's a bad call that shouldn't have been generated
    by the linker, so we may as well as just treat it like R_ARM_CALL.

    - Class: Static
    - Type: ARM (R_ARM_CALL, R_ARM_JUMP24); Deprecated (R_ARM_PC24)
    - Code: 1 (R_ARM_PC24), 28 (R_ARM_CALL), 29 (R_ARM_JUMP24)
    - Operation: ((S + A) | T) - P
      - S is the address of the symbol
      - A is the addend
      - P is the target location (place being relocated)
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction
    """

    @property
    def value(self):
        P = self.rebased_addr  # Location of this instruction
        A = inst = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)

        if inst & 0x00800000:
            A |= 0xFF000000  # Sign extend to 32-bits
        result = ((S + (A << 2)) | T) - P  # Do the initial work
        imm24 = (result & 0x03FFFFFE) >> 2  # Sign_extend(inst[25:2])

        if T:  # Do Thumb relocation
            mask = 0xFF000000
            bit_h = (result & 0x02) >> 1
            result = _applyReloc(inst, (0xFA | bit_h), mask)
        else:  # Do ARM relocation
            mask = 0xFFFFFF
            result = _applyReloc(inst, imm24, mask)

        log.debug("%s relocated as R_ARM_CALL with new instruction: %#x", self.symbol.name, result)
        return result


class R_ARM_PREL31(ELFReloc):
    """
    Relocate R_ARM_PREL31 symbols via instruction modification. The difference
    between this and R_ARM_CALL/R_ARM_PC24/R_ARM_JUMP24 is that it's a data
    relocation

    - Class: Static
    - Type: Data
    - Code: 42
    - Operation: ((S + A) | T) - P
      - S is the address of the symbol
      - A is the addend
      - P is the target location (place being relocated)
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction
    """

    @property
    def value(self):
        P = self.rebased_addr  # Location of this instruction
        A = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)

        if A & 0x01000000:
            A |= 0xF1000000  # Sign extend 31-bits
        result = ((S + A) | T) - P  # Do the initial work
        mask = 0x7FFFFFFF
        rel31 = result & mask
        result = _applyReloc(A, rel31, mask)
        log.debug("%s relocated as R_ARM_PREL31 to: 0x%x", self.symbol.name, result)
        return result


class R_ARM_REL32(ELFReloc):
    """
    Relocate R_ARM_REL32 symbols. This is essentially the same as
    GenericPCRelativeAddendReloc with the addition of a check
    for whether or not the target is Thumb.

    - Class: Static
    - Type: Data
    - Code: 3
    - Operation: ((S + A) | T) - P
      - S is the address of the symbol
      - A is the addend
      - P is the target location (place being relocated)
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction
    """

    @property
    def value(self):
        P = self.rebased_addr  # Location of this instruction
        A = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)
        result = ((S + A) | T) - P
        log.debug("%s relocated as R_ARM_REL32 to: 0x%x", self.symbol.name, result)
        return result


class R_ARM_ABS32(ELFReloc):
    """
    Relocate R_ARM_ABS32 symbols. This is essentially the same as
    GenericAbsoluteAddendReloc with the addition of a check
    for whether or not the target is Thumb.

    - Class: Static
    - Type: Data
    - Code: 3
    - Operation: (S + A) | T
      - S is the address of the symbol
      - A is the addend
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction
    """

    @property
    def value(self):
        A = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)
        result = (S + A) | T
        log.debug("%s relocated as R_ARM_ABS32 to: 0x%x", self.symbol.name, result)
        return result


class R_ARM_MOVW_ABS_NC(ELFReloc):
    """
    Relocate R_ARM_MOVW_ABS_NC symbols.

    - Class: Static
    - Type: Instruction
    - Code: 43
    - Operation: (S + A) | T
      - S is the address of the symbol
      - A is the addend
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction
    """

    @property
    def value(self):
        inst = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)
        # initial addend is formed by interpreting the 16-bit literal field
        # of the instruction as a signed value
        A = ((inst & 0xF0000) >> 4) | (inst & 0xFFF)
        if A & 0x8000:
            # two's complement
            A = -((A ^ 0xFFFF) + 1)
        X = (S + A) | T
        MaskX = X & 0xFFFF
        # inst modification:
        part1 = MaskX >> 12
        part2 = MaskX & 0xFFF
        inst &= 0xFFF0F000  # clears inst[11, 0] and inst[19, 16]
        inst |= (part1 << 16) & 0xF0000  # inst[19, 16] = part1
        inst |= part2 & 0xFFF  # inst[11, 0] = part2
        log.debug("%s relocated as R_ARM_MOVW_ABS_NC to: 0x%x", self.symbol.name, inst)
        return inst


class R_ARM_MOVT_ABS(ELFReloc):
    """
    Relocate R_ARM_MOVT_ABS symbols.

    - Class: Static
    - Type: Instruction
    - Code: 44
    - Operation: S + A
      - S is the address of the symbol
      - A is the addend
    """

    @property
    def value(self):
        inst = self.addend  # The instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        # initial addend is formed by interpreting the 16-bit literal field
        # of the instruction as a signed value
        A = ((inst & 0xF0000) >> 4) | (inst & 0xFFF)
        if A & 0x8000:
            # two's complement
            A = -((A ^ 0xFFFF) + 1)
        X = S + A
        MaskX = X & 0xFFFF0000
        # inst modification:
        part1 = (MaskX >> 16) >> 12
        part2 = (MaskX >> 16) & 0xFFF
        inst &= 0xFFF0F000  # clears inst[11, 0] and inst[19, 16]
        inst |= (part1 << 16) & 0xF0000  # inst[19, 16] = part1
        inst |= part2 & 0xFFF  # inst[11, 0] = part2
        log.debug("%s relocated as R_ARM_MOVT_ABS to: 0x%x", self.symbol.name, inst)
        return inst


class R_ARM_THM_CALL(ELFReloc):
    """
    Relocate R_ARM_THM_CALL symbols via instruction modification.

    - Class: Static
    - Type: ARM (R_ARM_THM_CALL)
    - Code: 10
    - Operation: ((S + A) | T) - P

      - S is the address of the symbol
      - A is the addend
      - P is the target location (place being relocated)
      - T is 1 if the symbol is of type STT_FUNC and addresses a Thumb instruction (This bit is entirely irrelevant
        because the 1-bit of the address gets shifted off in the encoding)

    - Encoding: See http://hermes.wings.cs.wisc.edu/files/Thumb-2SupplementReferenceManual.pdf

      - Page 71 (3-31) has the chart
      - It appears that it mistakenly references the I1 and I2 bits as J1 and J2 in the chart (see the notes at the
        bottom of the page -- the ranges don't make sense)

      - However, the J1/J2 bits are XORed with !S bit in this case (see vex implementation:
        https://github.com/angr/vex/blob/6d1252c7ce8fe8376318b8f8bb8034058454c841/priv/guest_arm_toIR.c#L19219 )

      - Implementation appears correct with the bits placed into offset[23:22]
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._insn_bytes = None

    def resolve_symbol(self, solist, **kwargs):
        kwargs["thumb"] = True
        super().resolve_symbol(solist, **kwargs)

    @property
    def value(self):
        P = self.rebased_addr  # Location of this instruction
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        T = _isThumbFunc(self.symbol, S)
        A = 0

        # Deconstruct the instruction:
        #  Because this 4-byte instruction is treated as two 2-byte instructions,
        #  the bytes are in the order `b3 b4 b1 b2`, where b4 is the most significant.

        if self._insn_bytes is None:
            self._insn_bytes = self.owner.memory.load(self.relative_addr, 4)

        hi = (self._insn_bytes[1] << 8) | self._insn_bytes[0]
        lo = (self._insn_bytes[3] << 8) | self._insn_bytes[2]
        inst = (hi << 16) | lo

        def gen_mask(n_bits, first_bit):
            """
            Builds a mask that captures n_bits, where the first bit captured is first_bit
            """
            return ((1 << n_bits) - 1) << first_bit

        if self.is_rela:
            A = self.addend
        else:
            # Build A (the initial addend)

            A |= (inst & gen_mask(11, 0)) << 1  # A[11:1]  = inst[10:0] (inclusive)
            A |= ((inst & gen_mask(10, 16)) >> 16) << 12  # A[21:12] = inst[25:16]

            sign_bit = bool(inst & gen_mask(1, 26)) & 1  # sign_bit = inst[26]

            J1 = (bool(inst & gen_mask(1, 13)) & 1) ^ (not sign_bit)  # J1 = inst[13] ^ !sign
            J2 = (bool(inst & gen_mask(1, 11)) & 1) ^ (not sign_bit)  # J2 = inst[11] ^ !sign

            A |= J1 << 23  # A[23] = J1
            A |= J2 << 22  # A[22] = J2

            A &= 0x7FFFFF

            if sign_bit:
                A |= 0xFF800000

        # Compute X, the new offset, from the symbol addr, S, the addend, A,
        #  the thumb flag, T, and PC, P.

        x = (((S + A) | T) - P) & 0xFFFFFFFF  # Also mask to 32 bits

        # Ensure jump is in range

        if x & 0xFF800000 != 0 and x & 0xFF800000 != 0xFF800000:
            raise CLEOperationError(
                "Jump target out of range for reloc R_ARM_THM_CALL (+- 2^23). "
                "This may be due to SimProcedures being allocated outside the jump range. "
                "If you believe this is the case, set 'rebase_granularity'=0x1000 in the "
                "load options."
            )

        # Rebuild the instruction, first clearing out any previously set offset bits

        #                 offset     1 2  offset
        #          11110S [21:12]  11J?J  [11:1]     (if ? is 1, BL; if ? is 0, BLX)
        inst &= ~0b00000111111111110010111111111111
        #         |       |       |       |       |
        #        32      24      16       8       0

        sign_bit = bool(x & gen_mask(1, 24)) & 1
        J1 = (bool(x & gen_mask(1, 23)) & 1) ^ (not sign_bit)
        J2 = (bool(x & gen_mask(1, 22)) & 1) ^ (not sign_bit)

        inst |= sign_bit << 26
        inst |= J1 << 13
        inst |= J2 << 11

        inst |= (x & gen_mask(11, 1)) >> 1
        inst |= ((x & gen_mask(10, 12)) >> 12) << 16

        # Put it back into <little endian short> <little endian short> format

        raw = ((inst & 0x00FF0000) >> 16, (inst & 0xFF000000) >> 24, (inst & 0x00FF), (inst & 0xFF00) >> 8)

        # The relocation handler expects a little-endian result, so flip it around.

        result = (raw[3] << 24) | (raw[2] << 16) | (raw[1] << 8) | raw[0]

        log.debug("%s relocated as R_ARM_THM_CALL with new instruction: %#x", self.symbol.name, result)
        return result


class R_ARM_COPY(GenericCopyReloc):
    pass


class R_ARM_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_ARM_JUMP_SLOT(GenericJumpslotReloc):
    pass


class R_ARM_RELATIVE(GenericRelativeReloc):
    pass


class R_ARM_ABS32_NOI(GenericAbsoluteAddendReloc):
    pass


class R_ARM_REL32_NOI(GenericPCRelativeAddendReloc):
    pass


class R_ARM_TLS_DTPMOD32(GenericTLSModIdReloc):
    pass


class R_ARM_TLS_DTPOFF32(GenericTLSDoffsetReloc):
    pass


class R_ARM_TLS_TPOFF32(GenericTLSOffsetReloc):
    pass


class R_ARM_JUMP24(R_ARM_CALL):
    pass


class R_ARM_PC24(R_ARM_CALL):
    pass


# EDG says: Implementing these the easy way.
# Inaccuracies may exist.  This is ARM, after all.
class R_ARM_THM_JUMP24(R_ARM_THM_CALL):
    pass


class R_ARM_THM_JUMP19(R_ARM_THM_CALL):
    pass


class R_ARM_THM_JUMP6(R_ARM_THM_CALL):
    pass


class R_ARM_THM_MOVW_ABS_NC(ELFReloc):
    """
    ((S + A) | T) & 0xffff
    Ref: https://github.com/ARM-software/abi-aa/blob/main/aaelf32/aaelf32.rst
    """

    @property
    def value(self):
        insn_bytes = self.owner.memory.load(self.relative_addr, 4)
        hi = (insn_bytes[1] << 8) | insn_bytes[0]
        lo = (insn_bytes[3] << 8) | insn_bytes[2]
        inst = (hi << 16) | lo
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        # initial addend is formed by interpreting the 16-bit literal field
        # of the instruction as a signed value
        A = (inst & 0b0000_0100_0000_0000_0000_0000_0000_0000) >> 26 << 15
        A |= (inst & 0b0000_0000_0000_1111_0000_0000_0000_0000) >> 16 << 11
        A |= (inst & 0b0000_0000_0000_0000_0111_0000_0000_0000) >> 12 << 8
        A |= inst & 0b0000_0000_0000_0000_0000_0000_1111_1111
        if A & 0x8000:
            # two's complement
            A = -((A ^ 0xFFFF) + 1)
        T = _isThumbFunc(self.symbol, S)
        X = (S + A) | T
        MaskX = X & 0xFFFF
        # inst modification:
        part1 = MaskX >> 12  # [19:16]
        part2 = (MaskX >> 11) & 0x1  # [26]
        part3 = (MaskX >> 8) & 0x7  # [14:12]
        part4 = MaskX & 0xFF  # [7:0]
        inst &= 0b1111_1011_1111_0000_1000_1111_0000_0000
        inst |= (part1 << 16) & 0b0000_0000_0000_1111_0000_0000_0000_0000
        inst |= (part2 << 26) & 0b0000_0100_0000_0000_0000_0000_0000_0000
        inst |= (part3 << 12) & 0b0000_0000_0000_0000_0111_0000_0000_0000
        inst |= (part4 << 0) & 0b0000_0000_0000_0000_0000_0000_1111_1111
        raw = ((inst & 0x00FF0000) >> 16, (inst & 0xFF000000) >> 24, (inst & 0x00FF), (inst & 0xFF00) >> 8)
        inst = (raw[3] << 24) | (raw[2] << 16) | (raw[1] << 8) | raw[0]
        log.debug("%s relocated as R_ARM_THM_MOVW_ABS_NC to: 0x%x", self.symbol.name, inst)
        return inst


class R_ARM_THM_MOVT_ABS(ELFReloc):
    """
    (S + A) & 0xffff0000
    Ref: https://github.com/ARM-software/abi-aa/blob/main/aaelf32/aaelf32.rst
    """

    @property
    def value(self):
        insn_bytes = self.owner.memory.load(self.relative_addr, 4)
        hi = (insn_bytes[1] << 8) | insn_bytes[0]
        lo = (insn_bytes[3] << 8) | insn_bytes[2]
        inst = (hi << 16) | lo
        S = self.resolvedby.rebased_addr  # The symbol's "value", where it points to
        # initial addend is formed by interpreting the 16-bit literal field
        # of the instruction as a signed value
        A = (inst & 0b0000_0100_0000_0000_0000_0000_0000_0000) >> 26 << 15
        A |= (inst & 0b0000_0000_0000_1111_0000_0000_0000_0000) >> 16 << 11
        A |= (inst & 0b0000_0000_0000_0000_0111_0000_0000_0000) >> 12 << 8
        A |= inst & 0b0000_0000_0000_0000_0000_0000_1111_1111
        if A & 0x8000:
            # two's complement
            A = -((A ^ 0xFFFF) + 1)
        X = S + A
        MaskX = X & 0xFFFF0000
        # inst modification:
        part1 = MaskX >> 28  # [19:16]
        part2 = (MaskX >> 27) & 0x1  # [26]
        part3 = (MaskX >> 24) & 0x7  # [14:12]
        part4 = (MaskX >> 16) & 0xFF  # [7:0]
        inst &= 0b1111_1011_1111_0000_1000_1111_0000_0000
        inst |= (part1 << 16) & 0b0000_0000_0000_1111_0000_0000_0000_0000
        inst |= (part2 << 26) & 0b0000_0100_0000_0000_0000_0000_0000_0000
        inst |= (part3 << 12) & 0b0000_0000_0000_0000_0111_0000_0000_0000
        inst |= (part4 << 0) & 0b0000_0000_0000_0000_0000_0000_1111_1111
        raw = ((inst & 0x00FF0000) >> 16, (inst & 0xFF000000) >> 24, (inst & 0x00FF), (inst & 0xFF00) >> 8)
        inst = (raw[3] << 24) | (raw[2] << 16) | (raw[1] << 8) | raw[0]
        log.debug("%s relocated as R_ARM_THM_MOVT_ABS to: 0x%x", self.symbol.name, inst)
        return inst


class R_ARM_GOT_PREL(GenericPCRelativeAddendReloc, RelocTruncate32Mixin, RelocGOTMixin):
    """
    GOT(S) + A - P
    Ref: https://github.com/ARM-software/abi-aa/blob/main/aaelf32/aaelf32.rst
    """


relocation_table_arm = {
    1: R_ARM_PC24,
    2: R_ARM_ABS32,
    3: R_ARM_REL32,
    # 4: ARM_LDR_PC_G0,
    # 5: R_ARM_ABS16,
    # 6: R_ARM_ABS12,
    # 7: R_ARM_THM_ABS5,
    # 8: R_ARM_ABS8,
    # 9: R_ARM_SBREL32,
    10: R_ARM_THM_CALL,
    # 11: R_ARM_THM_PC8,
    # 12: R_ARM_BREL_ADJ,
    # 13: R_ARM_TLS_DESC,
    # 14: R_ARM_THM_SWI8,
    # 15: R_ARM_XPC25,
    # 16: R_ARM_THM_XPC22,
    17: R_ARM_TLS_DTPMOD32,
    18: R_ARM_TLS_DTPOFF32,
    19: R_ARM_TLS_TPOFF32,
    20: R_ARM_COPY,
    21: R_ARM_GLOB_DAT,
    22: R_ARM_JUMP_SLOT,
    23: R_ARM_RELATIVE,
    # 24: R_ARM_GOTOFF,
    # 25: R_ARM_BASE_PREL,
    # 26: R_ARM_GOT_BREL,
    # 27: R_ARM_PLT32,
    28: R_ARM_CALL,
    29: R_ARM_JUMP24,
    30: R_ARM_THM_JUMP24,
    # 31: R_ARM_BASE_ABS,
    # 32: R_ARM_ALU_PCREL_7_0,
    # 33: R_ARM_ALU_PCREL_15_8,
    # 34: R_ARM_ALU_PCREL_23_15,
    # 35: R_ARM_LDR_SBREL_11_0_NC,
    # 36: R_ARM_ALU_SBREL_19_12_NC,
    # 37: R_ARM_ALU_SBREL_27_20_CK,
    # 38: R_ARM_TARGET1,
    # 39: R_ARM_SBREL31,
    # 40: R_ARM_V4BX,
    # 41: R_ARM_TARGET2,
    42: R_ARM_PREL31,
    43: R_ARM_MOVW_ABS_NC,
    44: R_ARM_MOVT_ABS,
    # 45: R_ARM_MOVW_PREL_NC,
    # 46: R_ARM_MOVT_PREL,
    47: R_ARM_THM_MOVW_ABS_NC,
    48: R_ARM_THM_MOVT_ABS,
    # 49: R_ARM_THM_MOVW_PREL_NC,
    # 50: R_ARM_THM_MOVT_PREL,
    51: R_ARM_THM_JUMP19,
    52: R_ARM_THM_JUMP6,
    # 53: R_ARM_THM_ALU_PREL_11_0,
    # 54: R_ARM_THM_PC12,
    55: R_ARM_ABS32_NOI,
    56: R_ARM_REL32_NOI,
    # 57: R_ARM_ALU_PC_G0_NC,
    # 58: R_ARM_ALU_PC_G0,
    # 59: R_ARM_ALU_PC_G1_NC,
    # 60: R_ARM_ALU_PC_G1,
    # 61: R_ARM_ALU_PC_G2,
    # 62: R_ARM_LDR_PC_G1,
    # 63: R_ARM_LDR_PC_G2,
    # 64: R_ARM_LDRS_PC_G0,
    # 65: R_ARM_LDRS_PC_G1,
    # 66: R_ARM_LDRS_PC_G2,
    # 67: R_ARM_LDC_PC_G0,
    # 68: R_ARM_LDC_PC_G1,
    # 69: R_ARM_LDC_PC_G2,
    # 70: R_ARM_ALU_SB_G0_NC,
    # 71: R_ARM_ALU_SB_G0,
    # 72: R_ARM_ALU_SB_G1_NC,
    # 73: R_ARM_ALU_SB_G1,
    # 74: R_ARM_ALU_SB_G2,
    # 75: R_ARM_LDR_SB_G0,
    # 76: R_ARM_LDR_SB_G1,
    # 77: R_ARM_LDR_SB_G2,
    # 78: R_ARM_LDRS_SB_G0,
    # 79: R_ARM_LDRS_SB_G1,
    # 80: R_ARM_LDRS_SB_G2,
    # 81: R_ARM_LDC_SB_G0,
    # 82: R_ARM_LDC_SB_G1,
    # 83: R_ARM_LDC_SB_G2,
    # 84: R_ARM_MOVW_BREL_NC,
    # 85: R_ARM_MOVT_BREL,
    # 86: R_ARM_MOVW_BREL,
    # 87: R_ARM_THM_MOVW_BREL_NC,
    # 88: R_ARM_THM_MOVT_BREL,
    # 89: R_ARM_THM_MOVW_BREL,
    # 90: R_ARM_TLS_GOTDESC,
    # 91: R_ARM_TLS_CALL,
    # 92: R_ARM_TLS_DESCSEQ,
    # 93: R_ARM_THM_TLS_CALL,
    # 94: R_ARM_PLT32_ABS,
    # 95: R_ARM_GOT_ABS,
    96: R_ARM_GOT_PREL,
    # 97: R_ARM_GOT_BREL12,
    # 98: R_ARM_GOTOFF12,
    # 99: R_ARM_GOTRELAX,
    # 100: R_ARM_GNU_VTENTRY,
    # 101: R_ARM_GNU_VTINHERIT,
    # 102: R_ARM_JUMP11,
    # 103: R_ARM_THM_PC9,
    # 104: R_ARM_TLS_GD32,
    # 105: R_ARM_TLS_LDM32,
    # 106: R_ARM_TLS_LDO32,
    # 107: R_ARM_TLS_IE32,
    # 108: R_ARM_TLS_LE32,
    # 109: R_ARM_TLS_LDO12,
    # 110: R_ARM_TLS_LE12,
    # 111: R_ARM_TLS_IE12GP,
    # 112-127: R_ARM_PRIVATE_<n>
    # 128: R_ARM_ME_TOO,
    # 129: R_ARM_THM_TLS_DESCSEQ16,
    # 130: R_ARM_THM_TLS_DESCSEQ32,
    # 131: R_ARM_THM_GOT_BREL12,
    # 132: R_ARM_THM_ALU_ABS_G0_NC,
    # 133: R_ARM_THM_ALU_ABS_G1_NC,
    # 134: R_ARM_THM_ALU_ABS_G2_NC,
    # 135: R_ARM_THM_ALU_ABS_G3,
    # 136: R_ARM_THM_BF16,
    # 137: R_ARM_THM_BF12,
    # 138: R_ARM_THM_BF18,
}


__all__ = ("relocation_table_arm",)



from __future__ import annotations

from cle.backends.symbol import SymbolSubType, SymbolType


class ELFSymbolType(SymbolSubType):
    """
    ELF-specific symbol types
    """

    # Enum classes cannot be inherited. Therefore, additional platform-specific
    # values should simply be added to this enumeration (e.g., STT_GNU_IFUNC)
    # with an appropriate conversion in `to_base_type()`.
    #
    # Though that could be solved with IntEnum as well, that breaks the
    # strong typing and is discouraged by Python docs.

    # Basic types
    STT_NOTYPE = (0, None)  # Symbol's type is not specified
    STT_OBJECT = (1, None)  # Symbol is a data object (variable, array, etc.)
    STT_FUNC = (2, None)  # Symbol is executable code (function, etc.)
    STT_SECTION = (3, None)  # Symbol refers to a section
    STT_FILE = (4, None)  # Local, absolute symbol that refers to a file
    STT_COMMON = (5, None)  # An uninitialized common block
    STT_TLS = (6, None)  # Thread local data object

    # ELF's generic place-holders
    STT_LOOS = (10, None)  # Lowest operating system-specific symbol type
    STT_HIOS = (12, None)  # Highest operating system-specific symbol type

    STT_LOPROC = (13, None)  # Lowest processor-specific symbol type
    STT_HIPROC = (15, None)  # Highest processor-specific symbol type

    #
    # OS- and processor-specific types. Note that the entire range
    # of values is used indiscriminantly for OS or processor.
    #
    # Try to use values that map to an `archinfo.Arch` so that `arch_from_id()`
    # is able to return a specific type. Otherwise, use something indicative
    # of its purpose.
    #

    # GNU indirect function
    #
    # HACK: It's GNU-specific, not OS-specific but GNU doesn't care. This
    # shouldn't be an issue unless someone tries analyzing an old ELF that
    # uses STT_LOOS for something else, before STT_GNU_IFUNC came about, in
    # which case angr will need a new SimOS variant anyway.
    STT_GNU_IFUNC = (STT_LOOS[0], "gnu")

    #
    # Below are examples of additional types that can be added. These are
    # commented out since they've never actually been used or tested.
    #

    # AMDGPU HSA
    #
    # https://github.com/RadeonOpenCompute/ROCR-Runtime/blob/master/src/inc/amd_hsa_elf.h
    # TODO: Update the arch name here if this arch is ever supported
    # STT_AMDGPU_HSA_KERNEL = (STT_LOOS[0], 'amdgpu_hsa')
    # STT_AMDGPU_HSA_INDIRECT_FUNCTION = (STT_LOOS[0] + 1, 'amdgpu_hsa')
    # STT_AMDGPU_HSA_METADATA = (STT_LOOS[0] + 2, 'amdgpu_hsa')

    # HP Precision Architecture (PA-RISC)
    #
    # https://github.com/lattera/glibc/blob/master/elf/elf.h
    # TODO: Update the arch name here if this arch is ever supported
    # STT_HP_OPAQUE = (STT_LOOS[0] + 1, 'hppa')
    # STT_HP_STUB = (STT_LOOS[0] + 2, 'hppa')
    # STT_PARISC_MILLICODE = (STT_LOPROC[0], 'hppa')

    def __init__(self, *args):  # pylint: disable=unused-argument
        # Essentially a static type check, this will fail on import
        # if someone defines a type that's not a `tuple`
        if not isinstance(self.value, tuple):
            raise ValueError(
                f"Symbol value '{self.value}' for member '{self.name}' is invalid. Values must be tuples."
            )  # pylint: disable=logging-format-interpolation

    def __repr__(self):
        return f"ELFSymbolType.{self.name}: (elf_value: {self.elf_value}, os_proc: {self.os_proc})"

    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        return self.value[0] == other.value[0]

    def __ne__(self, other):
        return not (self == other)

    @property
    def elf_value(self):
        return self.value[0]  # pylint: disable=unsubscriptable-object

    @property
    def os_proc(self):
        return self.value[1]  # pylint: disable=unsubscriptable-object

    @property
    def is_custom_os_proc(self):
        if self.elf_value in range(self.STT_LOOS.elf_value, self.STT_HIPROC.elf_value + 1):  # pylint: disable=no-member
            return self.os_proc is not None
        return False

    def to_base_type(self):
        if self is ELFSymbolType.STT_NOTYPE:
            return SymbolType.TYPE_NONE

        elif self in [ELFSymbolType.STT_FUNC, ELFSymbolType.STT_GNU_IFUNC]:
            return SymbolType.TYPE_FUNCTION

        elif self in [ELFSymbolType.STT_OBJECT, ELFSymbolType.STT_COMMON]:
            return SymbolType.TYPE_OBJECT

        elif self is ELFSymbolType.STT_SECTION:
            return SymbolType.TYPE_SECTION

        elif self is ELFSymbolType.STT_TLS:
            return SymbolType.TYPE_TLS_OBJECT

        elif self is ELFSymbolType.STT_GNU_IFUNC:
            return SymbolType.TYPE_FUNCTION

        else:
            return SymbolType.TYPE_OTHER


def __ELFSymbolTypeArchParser(cls, value):
    """
    This is just a nice way to allow for just specifying the `int` for
    default types: `ELFSymbolType(10)` rather than `ELFSymbolType((10,None))`.

    Idea courtesy: https://stackoverflow.com/q/24105268/1137728.

    We don't need to implement the `str` parsing like the SO link above since
    `Enum` already has built-in item access: `ELFSymbolType['STT_FUNC']`.
    """
    if isinstance(value, int):
        return super(ELFSymbolType, cls).__new__(cls, (value, None))
    else:
        return super(ELFSymbolType, cls).__new__(cls, value)


setattr(ELFSymbolType, "__new__", __ELFSymbolTypeArchParser)



from __future__ import annotations

import logging

import archinfo

from cle.address_translator import AT
from cle.errors import CLEError

from .backend import Backend, register_backend
from .relocation import Relocation
from .symbol import Symbol, SymbolType

log = logging.getLogger(name=__name__)

try:
    import binaryninja as bn
except ImportError:
    bn = None
    log.debug("Unable to import binaryninja module")
    BINJA_NOT_INSTALLED_STR = (
        "Binary Ninja does not appear to be installed. Please ensure Binary Ninja "
        "and its Python API are properly installed before using this backend."
    )


class BinjaSymbol(Symbol):
    BINJA_FUNC_SYM_TYPES = (
        [bn.SymbolType.ImportedFunctionSymbol, bn.SymbolType.FunctionSymbol, bn.SymbolType.ImportAddressSymbol]
        if bn
        else []
    )

    BINJA_DATA_SYM_TYPES = [bn.SymbolType.ImportedDataSymbol, bn.SymbolType.DataSymbol] if bn else []

    BINJA_IMPORT_TYPES = (
        [bn.SymbolType.ImportedFunctionSymbol, bn.SymbolType.ImportAddressSymbol, bn.SymbolType.ImportedDataSymbol]
        if bn
        else []
    )

    def __init__(self, owner, sym):
        if not bn:
            raise CLEError(BINJA_NOT_INSTALLED_STR)

        if sym.type in self.BINJA_FUNC_SYM_TYPES:
            symtype = SymbolType.TYPE_FUNCTION
        elif sym.type in self.BINJA_DATA_SYM_TYPES:
            symtype = SymbolType.TYPE_OBJECT
        else:
            symtype = SymbolType.TYPE_OTHER

        super().__init__(owner, sym.raw_name, AT.from_rva(sym.address, owner).to_rva(), owner.bv.address_size, symtype)

        if sym.type in self.BINJA_IMPORT_TYPES:
            self.is_import = True

        # TODO: set is_weak appropriately


class BinjaReloc(Relocation):
    @property
    def value(self):
        return self.relative_addr


class BinjaBin(Backend):
    """
    Get information from binaries using Binary Ninja. Basing this on idabin.py, but will try to be more complete.
    TODO: add more features as Binary Ninja's feature set improves
    """

    is_default = True  # Tell CLE to automatically consider using the BinjaBin backend
    BINJA_ARCH_MAP = {
        "aarch64": archinfo.ArchAArch64(endness="Iend_LE"),
        "armv7": archinfo.ArchARMEL(endness="Iend_LE"),
        "thumb2": archinfo.ArchARMEL(endness="Iend_LE"),
        "armv7eb": archinfo.ArchARMEL(endness="Iend_BE"),
        "thumb2eb": archinfo.ArchARMEL(endness="Iend_BE"),
        "mipsel32": archinfo.ArchMIPS32(endness="Iend_LE"),
        "mips32": archinfo.ArchMIPS32(endness="Iend_BE"),
        "ppc": archinfo.ArchPPC32(endness="Iend_BE"),
        "ppc_le": archinfo.ArchPPC32(endness="Iend_LE"),
        "x86": archinfo.ArchX86(),
        "x86_64": archinfo.ArchAMD64(),
    }

    def __init__(self, binary, *args, **kwargs):
        super().__init__(binary, *args, **kwargs)
        if not bn:
            raise CLEError(BINJA_NOT_INSTALLED_STR)
        # get_view_of_file can take a bndb or binary - wait for autoanalysis to complete
        self.bv = bn.BinaryViewType.get_view_of_file(binary, False)
        log.info("Analyzing %s, this may take some time...", binary)
        self.bv.update_analysis_and_wait()
        log.info("Analysis complete")
        # Note may want to add option to kick off linear sweep

        try:
            self.set_arch(self.BINJA_ARCH_MAP[self.bv.arch.name])
        except KeyError:
            log.error("Architecture %s is not supported.", self.bv.arch.name)

        for seg in self.bv.segments:
            log.info("Adding memory for segment at %x.", seg.start)
            br = bn.BinaryReader(self.bv)
            br.seek(seg.start)
            data = br.read(len(seg))
            self.memory.add_backer(seg.start, data)

        self._find_got()
        self._symbol_cache = {}
        self._init_symbol_cache()
        # Note: this represents the plt stub. ImportAddressSymbol refers to .got entries
        # Since we're not trying to import and load dependencies directly, but want to run SimProcedures,
        # We should use the binaryninja.SymbolType.ImportedFunctionSymbol
        # Also this should be generalized to get data imports, too
        self.raw_imports = {
            i.name: i.address for i in self.bv.get_symbols_of_type(bn.SymbolType.ImportedFunctionSymbol)
        }
        self._process_imports()
        self.exports = {}
        self.linking = "static" if len(self.raw_imports) == 0 else "dynamic"
        # We'll look for this attribute to see if we need to do SimProcedures for any imports in this binary
        # This is an ugly hack, but will have to use this for now until Binary Ninja exposes dependencies
        self.guess_simprocs = True
        self.guess_simprocs_hint = "nix" if self.bv.get_section_by_name(".plt") else "win"
        log.warning(
            "This backend is based on idabin.py.\n"
            "You may encounter unexpected behavior if:\n"
            "\tyour target depends on library data symbol imports, or\n"
            "\tlibrary imports that don't have a guess-able SimProcedure\n"
            "Good luck!"
        )

    def _process_imports(self):
        """Process self.raw_imports into list of Relocation objects"""
        if not self.raw_imports:
            log.warning("No imports found - if this is a dynamically-linked binary, something probably went wrong.")

        for name, addr in self.raw_imports.items():
            BinjaReloc(self, self._symbol_cache[name], addr)

    def _init_symbol_cache(self):
        # Note that we could also access name, short_name, or full_name attributes
        for sym in self.bv.get_symbols():
            cle_sym = BinjaSymbol(self, sym)
            self._symbol_cache[sym.raw_name] = cle_sym
            self.symbols.add(cle_sym)

    def _find_got(self):
        """
        Locate the section (e.g. .got) that should be updated when relocating functions (that's where we want to
        write absolute addresses).
        """
        sec_name = self.arch.got_section_name
        self.got_begin = None
        self.got_end = None

        try:
            got_sec = self.bv.sections[self.arch.got_section_name]
            self.got_begin = got_sec.start
            self.got_end = got_sec.end
        except KeyError:
            log.warning("No got section mapping found!")

        # If we reach this point, we should have the addresses
        if self.got_begin is None or self.got_end is None:
            log.warning("No section %s, is this a static binary ? (or stripped)", sec_name)
            return False
        return True

    @staticmethod
    def is_compatible(stream):
        if not bn:
            return False
        magic = stream.read(100)
        stream.seek(0)
        # bndb files are SQlite 3
        if magic.startswith(b"SQLite format 3") and stream.name.endswith("bndb"):
            return True

        return False

    def in_which_segment(self, addr):
        """
        Return the segment name at address `addr`.
        """
        # WARNING: if there are overlapping sections, we choose the first name.
        # The only scenario I've seen here is a NOBITS section that "overlaps" with another one, but
        # I'm not sure if that's a heurstic that should be applied here.
        # https://stackoverflow.com/questions/25501044/gcc-ld-overlapping-sections-tbss-init-array-in-statically-linked-elf-bin#25771838
        seg = self.bv.get_sections_at(addr)[0].name
        return "unknown" if len(seg) == 0 else seg

    def get_symbol_addr(self, sym):
        """
        Get the address of the symbol `sym` from IDA.

        :returns: An address.
        """
        # sym is assumed to be the raw_name of the symbol
        return self.bv.get_symbol_by_raw_name(sym)

    def function_name(self, addr):
        """
        Return the function name at address `addr`.
        """
        func = self.bv.get_function_at(addr)
        if not func:
            return "UNKNOWN"
        return func.name

    @property
    def min_addr(self):
        """
        Get the min address of the binary. (note: this is probably not "right")
        """
        return self.bv.start

    @property
    def max_addr(self):
        """
        Get the max address of the binary.
        """
        return self.bv.end

    @property
    def entry(self):
        if self._custom_entry_point is not None:
            return self._custom_entry_point + self.mapped_base
        return self.bv.entry_point + self.mapped_base

    def get_strings(self):
        """
        Extract strings from binary (Binary Ninja).

        :returns:   An array of strings.
        """
        return self.bv.get_strings()

    def set_got_entry(self, name, newaddr):
        """
        Resolve import `name` with address `newaddr`. That is, update the GOT entry for `name` with `newaddr`.
        """
        if name not in self.imports:
            log.warning("%s not in imports", name)
            return

        addr = self.imports[name]
        self.memory.pack_word(addr, newaddr)

    def close(self):
        """
        Release the BinaryView we created in __init__
        :return: None
        """
        self.bv.file.close()


register_backend("binja", BinjaBin)



from __future__ import annotations


class PatchedStream:
    """
    An object that wraps a readable stream, performing passthroughs on seek and read operations,
    except to make it seem like the data has actually been patched by the given patches.
    """

    def __init__(self, stream, patches):
        """
        :param stream:      The stream to patch
        :param patches:     A list of tuples of (addr, patch data)
        """
        if type(stream) is PatchedStream:
            patches = stream.patches + patches
            stream = stream.stream

        self.stream = stream
        self.patches = patches
        self._pos = stream.tell()

    def read(self, *args, **kwargs):
        data = self.stream.read(*args, **kwargs)
        pos = self._pos
        newpos = pos + len(data)

        # AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
        for addr, patch in self.patches:
            if (addr >= self._pos and addr < newpos) or (self._pos >= addr and self._pos < addr + len(patch)):
                inject_start = max(0, addr - pos)
                inject_end = min(addr - pos + len(patch), len(data))
                patch_start = max(0, pos - addr)
                patch_end = min(newpos - addr, len(patch))
                data = data[:inject_start] + patch[patch_start:patch_end] + data[inject_end:]

        self._pos = newpos
        return data

    def seek(self, *args, **kwargs):
        self.stream.seek(*args, **kwargs)
        self._pos = self.stream.tell()

    def tell(self):
        return self._pos

    def close(self):
        return self.stream.close()



from __future__ import annotations

import logging

from cle.errors import CLEError

from .backend import Backend, register_backend
from .region import Segment

log = logging.getLogger(name=__name__)

__all__ = ("Blob",)


class Blob(Backend):
    """
    Representation of a binary blob, i.e. an executable in an unknown file format.
    """

    is_default = True  # Tell CLE to automatically consider using the Blob backend

    def __init__(self, *args, offset=None, segments=None, **kwargs):
        """
        :param arch:   (required) an :class:`archinfo.Arch` for the binary blob.
        :param offset: Skip this many bytes from the beginning of the file.
        :param segments:      List of tuples describing how to map data into memory. Tuples
                              are of ``(file_offset, mem_addr, size)``.

        You can't specify both ``offset`` and ``segments``.
        """
        if "custom_offset" in kwargs:
            offset = kwargs.pop("custom_offset")
            log.critical("Deprecation warning: the custom_offset parameter has been renamed to offset")
        super().__init__(*args, **kwargs)
        self.set_load_args(offset=offset, segments=segments)

        if self._arch is None:
            raise CLEError("Must specify arch when loading blob!")

        if self._custom_entry_point is None:
            log.warning("No entry_point was specified for blob %s, assuming 0", self.binary_basename)

        self._entry = 0
        self._max_addr = 0
        self._min_addr = 2**64

        try:
            self.linked_base = kwargs["base_addr"]
        except KeyError:
            log.warning("No base_addr was specified for blob %s, assuming 0", self.binary_basename)
        self.mapped_base = self.linked_base

        self.os = "unknown"

        if offset is not None:
            if segments is not None:
                log.error("You can't specify both offset and segments. Taking only the segments data")
            else:
                self._binary_stream.seek(0, 2)
                segments = [(offset, self.linked_base, self._binary_stream.tell() - offset)]
        else:
            if segments is not None:
                pass
            else:
                self._binary_stream.seek(0, 2)
                segments = [(0, self.linked_base, self._binary_stream.tell())]

        for file_offset, mem_addr, size in segments:
            self._load(file_offset, mem_addr, size)

    @staticmethod
    def is_compatible(stream):
        return stream == 0  # I hate pylint

    @property
    def min_addr(self):
        return self._min_addr

    @property
    def max_addr(self):
        return self._max_addr

    def _load(self, file_offset, mem_addr, size):
        """
        Load a segment into memory.
        """

        self._binary_stream.seek(file_offset)
        string = self._binary_stream.read(size)
        if string:
            self.memory.add_backer(mem_addr - self.linked_base, string)
        seg = Segment(file_offset, mem_addr, size, size)
        self.segments.append(seg)
        self._max_addr = max(len(string) + mem_addr - 1, self._max_addr)
        self._min_addr = min(mem_addr, self._min_addr)

    def function_name(self, addr):  # pylint: disable=unused-argument,no-self-use
        """
        Blobs don't support function names.
        """
        return None

    def contains_addr(self, addr):
        return addr >= self.mapped_base and (addr - self.mapped_base) in self.memory

    def in_which_segment(self, addr):  # pylint: disable=unused-argument,no-self-use
        """
        Blobs don't support segments.
        """
        return None

    @classmethod
    def check_compatibility(cls, spec, obj):  # pylint: disable=unused-argument
        return True

    def _cache_content(self):
        return

    def _checksum(self):
        return


register_backend("blob", Blob)



from __future__ import annotations

from cle.backends.symbol import SymbolType

from .common import PointTo, StaticWord
from .simdata import SimData, register


class DummyProgname(SimData):
    name = "_dummy_progname"
    type = SymbolType.TYPE_OBJECT
    libname = "libc.so"

    progname = b"./program\0"

    @classmethod
    def static_size(cls, owner):
        return len(cls.progname)

    def value(self):
        return self.progname


class Progname(PointTo):
    pointto_name = "_dummy_progname"
    pointto_type = SymbolType.TYPE_OBJECT
    name = "__progname"
    libname = "libc.so"
    type = SymbolType.TYPE_OBJECT
    addend = 2


class PrognameFull(PointTo):
    pointto_name = "_dummy_progname"
    pointto_type = SymbolType.TYPE_OBJECT
    name = "__progname_full"
    libname = "libc.so.6"
    type = SymbolType.TYPE_OBJECT
    addend = 0


class EnvironmentPointer(StaticWord):
    name = "__environ"
    libname = "libc.so"
    word = 0


class EnvironmentPointerAlso(StaticWord):
    name = "environ"
    libname = "libc.so"
    word = 0


class OptInd(StaticWord):
    name = "optind"
    libname = "libc.so"
    word = 1
    wordsize = 4


class OptArg(StaticWord):
    name = "optarg"
    libname = "libc.so"
    word = 0


class Errno(StaticWord):
    type = SymbolType.TYPE_TLS_OBJECT
    name = "errno"
    libname = "libc.so"
    word = 0
    wordsize = 4


class LibcStackEnd(StaticWord):
    name = "__libc_stack_end"
    libname = "ld-linux"
    word = 0


class RTLDGlobal(SimData):
    name = "_rtld_global"
    type = SymbolType.TYPE_OBJECT
    libname = "ld-linux"

    @classmethod
    def static_size(cls, owner):
        return {"AMD64": 3960, "X86": 2100}.get(owner.arch.name, 1024)

    def value(self):
        return bytes(self.static_size(self.owner))


class RTLDGlobalRO(SimData):
    name = "_rtld_global_ro"
    type = SymbolType.TYPE_OBJECT
    libname = "ld-linux"

    @classmethod
    def static_size(cls, owner):
        return {
            "AMD64": 440,
            "X86": 576,
        }.get(owner.arch.name, 256)

    def value(self):
        return bytes(self.static_size(self.owner))


register(DummyProgname)
register(Progname)
register(PrognameFull)
register(EnvironmentPointer)
register(EnvironmentPointerAlso)
register(OptInd)
register(OptArg)
register(Errno)
register(LibcStackEnd)
register(RTLDGlobal)
register(RTLDGlobalRO)



from __future__ import annotations

import contextlib
import os

import elftools

from .errors import CLEError, CLEFileNotFoundError


# https://code.woboq.org/userspace/glibc/include/libc-pointer-arith.h.html#43
def ALIGN_DOWN(base, size):
    return base & -size


# https://code.woboq.org/userspace/glibc/include/libc-pointer-arith.h.html#50
def ALIGN_UP(base, size):
    return ALIGN_DOWN(base + size - 1, size)


# To verify the mmap behavior you can compile and run the following program. Fact is that mmap file mappings
# always map in the entire page into memory from the file if available. If not, it gets zero padded
# pylint: disable=pointless-string-statement
# #include <stdio.h>
# #include <sys/types.h>
# #include <sys/stat.h>
# #include <unistd.h>
# #include <fcntl.h>
# #include <sys/mman.h>
#
# void make_test_file()
# {
#     void* data = (void*)0xdead0000;
#     int fd = open("./test.data", O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
#     for (int i = 0; i < 0x1800; i += sizeof(void*)) // Only write 1 1/2 pages worth
#     {
#         write(fd, &data, sizeof(void*));
#         data += sizeof(void*);
#     }
#     close(fd);
# }
# int main(int argc, char* argv[])
# {
#     make_test_file();
#
#     int fd = open("./test.data", O_RDONLY);
#     unsigned char* mapping = mmap(NULL, 0x123, PROT_READ, MAP_PRIVATE, fd, 4096);
#
#     for (int i=0; i < 0x1000; i++)
#     {
#         printf("%02x ", mapping[i]);
#         if (i % sizeof(void*) == (sizeof(void*) - 1))
#             printf("| ");
#         if (i % 16 == 15)
#             printf("\n");
#     }
# }


def get_mmaped_data(stream, offset, length, page_size):
    if offset % page_size != 0:
        raise CLEError(
            "libc helper for mmap: Invalid page offset, should be multiple of page size! "
            f"Stream {stream}, offset {offset}, length: {length}"
        )

    read_length = ALIGN_UP(length, page_size)
    stream.seek(offset)
    data = stream.read(read_length)
    return data.ljust(read_length, b"\0")


@contextlib.contextmanager
def stream_or_path(obj, perms="rb"):
    if hasattr(obj, "read") and hasattr(obj, "seek"):
        obj.seek(0)
        yield obj
    else:
        if not os.path.exists(obj):
            raise CLEFileNotFoundError(f"{obj!r} is not a valid path")

        with open(obj, perms) as f:
            yield f


def key_bisect_floor_key(lst, key, lo=0, hi=None, keyfunc=lambda x: x):
    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(lst)
    while lo < hi:
        mid = (lo + hi) // 2
        if keyfunc(lst[mid]) <= key:
            lo = mid + 1
        else:
            hi = mid
    if lo <= len(lst) and lo > 0:
        return lst[lo - 1]
    return None


def key_bisect_find(lst, item, lo=0, hi=None, keyfunc=lambda x: x):
    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(lst)
    while lo < hi:
        mid = (lo + hi) // 2
        if keyfunc(lst[mid]) <= keyfunc(item):
            lo = mid + 1
        else:
            hi = mid
    return lo


def key_bisect_insort_left(lst, item, lo=0, hi=None, keyfunc=lambda x: x):
    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(lst)
    while lo < hi:
        mid = (lo + hi) // 2
        if keyfunc(lst[mid]) < keyfunc(item):
            lo = mid + 1
        else:
            hi = mid
    lst.insert(lo, item)


def key_bisect_insort_right(lst, item, lo=0, hi=None, keyfunc=lambda x: x):
    if lo < 0:
        raise ValueError("lo must be non-negative")
    if hi is None:
        hi = len(lst)
    while lo < hi:
        mid = (lo + hi) // 2
        if keyfunc(lst[mid]) <= keyfunc(item):
            lo = mid + 1
        else:
            hi = mid
    lst.insert(lo, item)


def get_text_offset(path):
    """
    Offset of text section in the binary.
    """
    with stream_or_path(path) as f:
        e = elftools.elf.elffile.ELFFile(f)
        return e.get_section_by_name(".text").header.sh_offset


def extract_null_terminated_bytestr(data: bytes | bytearray, offset: int = 0, sentinel_value: bytes = b"\x00") -> bytes:
    """
    Return an exclusive null-terminated sequence of bytes at `offset` in `data`.
    """
    return bytes(data[offset : data.find(sentinel_value, offset)])



from __future__ import annotations

import logging

from .amd64 import relocation_table_amd64
from .arm import relocation_table_arm
from .arm64 import relocation_table_arm64
from .i386 import relocation_table_i386
from .mips import relocation_table_mips
from .ppc import relocation_table_ppc
from .ppc64 import relocation_table_ppc64
from .s390x import relocation_table_s390x
from .sparc import relocation_table_sparc

ALL_RELOCATIONS = {
    "AMD64": relocation_table_amd64,
    "ARMCortexM": relocation_table_arm,
    "ARM": relocation_table_arm,
    "AARCH64": relocation_table_arm64,
    "ARMEL": relocation_table_arm,
    "ARMHF": relocation_table_arm,
    "X86": relocation_table_i386,
    "MIPS32": relocation_table_mips,
    "MIPS64": relocation_table_mips,
    "PPC32": relocation_table_ppc,
    "PPC64": relocation_table_ppc64,
    "S390X": relocation_table_s390x,
    "sparc:BE:32:default": relocation_table_sparc,
}


log = logging.getLogger(name=__name__)
complaint_log = set()


def get_relocation(arch, r_type):
    if r_type == 0:
        return None
    try:
        return ALL_RELOCATIONS[arch][r_type]
    except KeyError:
        if (arch, r_type) not in complaint_log:
            complaint_log.add((arch, r_type))
            log.warning("Unknown reloc %d on %s", r_type, arch)
        return None



from __future__ import annotations

from cle.memory import Clemory


class CryptSentinel(Clemory):
    """
    Mach-O binaries are often encrypted, and some area of memory is only decrypted at runtime later in the loading
    process. This decryption process can't easily be implemented in CLE and is typically done with separate tools
    But not all data is encrypted, and various metadata is still accessible.

    This Clemory serves as a shim that allows us to notice accesses to encrypted areas of memory and raise an exception
    This means that all code that was written will loudly fail on access to encrypted memory, instead of silently
    reading garbage data.
    """

    def __init__(self, arch, root=False):
        super().__init__(arch, root)
        self._crypt_start = None
        self._crypt_end = None
        self._is_encrypted: bool = False

    def load(self, addr, n):
        self._assert_unencrypted_access(addr, n)
        return super().load(addr, n)

    def store(self, addr, data):
        self._assert_unencrypted_access(addr, len(data))
        return super().store(addr, data)

    def find(self, data, search_min=None, search_max=None):
        if self._is_encrypted:
            raise EncryptedDataAccessException("Cannot search encrypted memory region", self._crypt_start)
        return super().find(data, search_min, search_max)

    def set_crypt_info(self, cryptid, start, size):
        self._is_encrypted = cryptid != 0 and size > 0
        self._crypt_start = start
        self._crypt_end = start + size

    def backers(self, addr=0):
        if self._is_encrypted:
            if self._crypt_start <= addr < self._crypt_end:
                raise EncryptedDataAccessException("Accessing encrypted memory region", addr)
        return super().backers(addr)

    def _assert_unencrypted_access(self, addr, size):
        """
        Make sure that the access does not cover encrypted memory regions
        If it does, raise an error

        Cases:
        - Access starts before encrypted region and ends after it
        - Access starts within encrypted region
        - Access ends within encrypted region

        :param addr:
        :param size:
        :return:
        """
        if not self._is_encrypted:
            return

        encrypted_range = range(self._crypt_start, self._crypt_end)
        if addr in encrypted_range or (addr + size) in encrypted_range or (addr < self._crypt_start < addr + size):
            raise EncryptedDataAccessException("Accessing encrypted memory region", addr)


class EncryptedDataAccessException(Exception):
    """
    Special exception to be raised when access to encrypted memory is attempted
    """

    def __init__(self, message, addr):
        super().__init__(message)
        self.addr = addr



from __future__ import annotations

import bisect
import struct
from mmap import mmap

import archinfo

__all__ = ("ClemoryBase", "Clemory", "ClemoryView", "ClemoryTranslator", "UninitializedClemory")


class ClemoryBase:
    """
    The base class of all Clemory classes.
    """

    __slots__ = ("_arch", "_pointer")

    def __init__(self, arch):
        self._arch = arch
        self._pointer = 0

    def __getitem__(self, k):
        raise NotImplementedError

    def __setitem__(self, k, v):
        raise NotImplementedError

    def __contains__(self, k):
        raise NotImplementedError

    def load(self, addr, n):
        raise NotImplementedError

    def store(self, addr, data):
        raise NotImplementedError

    def backers(self, addr=0):
        raise NotImplementedError

    def find(self, data, search_min=None, search_max=None):
        raise NotImplementedError

    def unpack(self, addr, fmt):
        """
        Use the ``struct`` module to unpack the data at address `addr` with the format `fmt`.
        """

        try:
            start, backer = next(self.backers(addr))
        except StopIteration:
            raise KeyError(addr)  # pylint: disable=raise-missing-from

        if start > addr:
            raise KeyError(addr)

        try:
            return struct.unpack_from(fmt, backer, addr - start)
        except struct.error as e:
            if len(backer) - (addr - start) >= struct.calcsize(fmt):
                raise e
            raise KeyError(addr)  # pylint: disable=raise-missing-from

    def unpack_word(self, addr, size=None, signed=False, endness=None):
        """
        Use the ``struct`` module to unpack a single integer from the address `addr`.

        You may override any of the attributes of the word being extracted:

        :param int size:    The size in bytes to pack/unpack. Defaults to wordsize (e.g. 4 bytes on
                            a 32 bit architecture)
        :param bool signed: Whether the data should be extracted signed/unsigned. Default unsigned
        :param archinfo.Endness endness: The endian to use in packing/unpacking. Defaults to memory endness
        """
        if size is not None and size > 8:
            # support larger wordsizes via recursive algorithm
            subsize = size >> 1
            if size != subsize << 1:
                raise ValueError("Cannot unpack non-power-of-two sizes")

            if endness is None:
                endness = self._arch.memory_endness
            if endness == archinfo.Endness.BE:
                lo_off, hi_off = subsize, 0
            elif endness == archinfo.Endness.LE:
                lo_off, hi_off = 0, subsize
            else:
                raise ValueError(f"Unsupported endness value {endness}.")

            lo = self.unpack_word(addr + lo_off, size=subsize, signed=False, endness=endness)
            hi = self.unpack_word(addr + hi_off, size=subsize, signed=signed, endness=endness)
            return (hi << (subsize << 3)) | lo

        return self.unpack(addr, self._arch.struct_fmt(size=size, signed=signed, endness=endness))[0]

    def load_null_terminated_bytes(self, addr, max_size=4096) -> bytes:
        """
        Load a null-terminated string from memory at address `addr` with a maximum size of `max_size`.
        Useful
        """
        data = bytearray()
        for i in range(max_size):
            try:
                byte = self[addr + i]
            except KeyError:
                break
            if byte == 0:
                break
            data.append(byte)
        return bytes(data)

    def pack(self, addr, fmt, *data):
        """
        Use the ``struct`` module to pack `data` into memory at address `addr` with the format `fmt`.
        """

        try:
            start, backer = next(self.backers(addr))
        except StopIteration:
            raise KeyError(addr)  # pylint: disable=raise-missing-from

        if start > addr:
            raise KeyError(addr)  # pylint: disable=raise-missing-from

        try:
            return struct.pack_into(fmt, backer, addr - start, *data)
        except struct.error as e:
            if len(backer) - (addr - start) >= struct.calcsize(fmt):
                raise e
            raise KeyError(addr)  # pylint: disable=raise-missing-from

    def pack_word(self, addr, data, size=None, signed=False, endness=None):
        """
        Use the ``struct`` module to pack a single integer `data` into memory at the address `addr`.

        You may override any of the attributes of the word being packed:

        :param int size:    The size in bytes to pack/unpack. Defaults to wordsize (e.g. 4 bytes on
                            a 32 bit architecture)
        :param bool signed: Whether the data should be extracted signed/unsigned. Default unsigned
        :param archinfo.Endness endness: The endian to use in packing/unpacking. Defaults to memory endness
        """
        if not signed:
            data &= (1 << (size * 8 if size is not None else self._arch.bits)) - 1
        return self.pack(addr, self._arch.struct_fmt(size=size, signed=signed, endness=endness), data)

    def read(self, nbytes):
        """
        The stream-like function that reads up to a number of bytes starting from the current
        position and updates the current position. Use with :func:`seek`.

        Up to `nbytes` bytes will be read, halting at the beginning of the first unmapped region
        encountered.
        """

        try:
            out = self.load(self._pointer, nbytes)
        except KeyError:
            return b""
        else:
            self._pointer += len(out)
            return out

    def seek(self, value):
        """
        The stream-like function that sets the "file's" current position. Use with :func:`read()`.

        :param value:        The position to seek to.
        """
        self._pointer = value

    def tell(self):
        return self._pointer

    def close(self):  # pylint: disable=no-self-use
        pass


class Clemory(ClemoryBase):
    """
    An object representing a memory space.

    Accesses can be made with [index] notation.
    """

    __slots__ = ("_backers", "_root", "consecutive", "min_addr", "max_addr")

    def __init__(self, arch, root=False):
        super().__init__(arch)
        self._backers: list[tuple[int, bytearray | Clemory | list[int]]] = []
        self._root = root
        self.consecutive = True
        self.min_addr = 0
        self.max_addr = 0

    def add_backer(self, start, data, overwrite=False):
        """
        Adds a backer to the memory.

        :param start:   The address where the backer should be loaded.
        :param data:    The backer itself. Can be either a bytestring or another :class:`Clemory`.
        :param overwrite:
                        If True and the range overlaps any existing backer, the existing backer will be split up and
                        the overlapping part will be replaced with the new backer.
        """
        if not data:
            raise ValueError("Backer is empty!")

        if not isinstance(data, bytes | bytearray | list | Clemory | mmap):
            raise TypeError("Data must be a bytes, list, or Clemory object.")
        if overwrite:
            if isinstance(data, Clemory):
                raise TypeError("Cannot perform an overwrite-add with a Clemory")
            self.split_backer(start)
            self.split_backer(start + len(data))
            try:
                self.remove_backer(start)
            except ValueError:
                pass
            try:
                existing, _ = next(self.backers(start + len(data)))
            except StopIteration:
                pass
            else:
                if existing < start + len(data):
                    self.remove_backer(existing)
        else:
            try:
                existing, _ = next(self.backers(start))
            except StopIteration:
                pass
            else:
                if existing <= start:
                    raise ValueError(f"Address {start:#x} is already backed!")
        if isinstance(data, Clemory) and data._root:
            raise ValueError("Cannot add a root clemory as a backer!")
        if isinstance(data, bytes):
            data = bytearray(data)
        bisect.insort(self._backers, (start, data), key=lambda x: x[0])
        self._update_min_max()

    def split_backer(self, addr):
        """
        Ensures that ``addr`` is the start of a backer, if it is backed.
        """
        try:
            start_addr, backer = next(self.backers(addr))
        except StopIteration:
            return
        if addr <= start_addr:
            return
        if isinstance(backer, ClemoryBase):
            raise ValueError("Cannot split a backer which is itself a clemory")
        if addr >= start_addr + len(backer):
            return

        self.remove_backer(start_addr)
        self.add_backer(start_addr, backer[: addr - start_addr])
        self.add_backer(addr, backer[addr - start_addr :])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} [{hex(self.min_addr)}:{hex(self.max_addr)}]>"

    def update_backer(self, start, data):
        if not isinstance(data, bytes | list | Clemory):
            raise TypeError("Data must be a bytes, list, or Clemory object.")
        if isinstance(data, bytes):
            data = bytearray(data)
        for i, (oldstart, _) in enumerate(self._backers):
            if oldstart == start:
                self._backers[i] = (start, data)
                break
        else:
            raise ValueError("Can't find backer to update")

        self._update_min_max()

    def remove_backer(self, start):
        for i, (oldstart, _) in enumerate(self._backers):
            if oldstart == start:
                self._backers.pop(i)
                break
        else:
            raise ValueError("Can't find backer to remove")

        self._update_min_max()

    def __iter__(self):
        for start, string in self._backers:
            if isinstance(string, bytes | list):
                for x in range(len(string)):
                    yield start + x
            else:
                for x in string:
                    yield start + x

    def __getitem__(self, k):
        for start, data in self._backers:
            if isinstance(data, bytearray | list):
                if 0 <= k - start < len(data):
                    return data[k - start]
            elif isinstance(data, Clemory):
                if data.min_addr <= k - start < data.max_addr:
                    try:
                        return data[k - start]
                    except KeyError:
                        pass
        raise KeyError(k)

    def __setitem__(self, k, v):
        for start, data in self._backers:
            if isinstance(data, bytearray | list):
                if 0 <= k - start < len(data):
                    data[k - start] = v
                    return
            elif isinstance(data, Clemory):
                if data.min_addr <= k - start < data.max_addr:
                    try:
                        data[k - start] = v
                        return
                    except KeyError:
                        pass
        raise KeyError(k)

    def __contains__(self, k):
        # Fast path
        if self.consecutive:
            return self.min_addr <= k < self.max_addr
        else:
            # Check if this is an empty Clemory instance
            if not self._backers:
                return False
            # Check if it is out of the memory range
            if k < self.min_addr or k >= self.max_addr:
                return False

        try:
            self.__getitem__(k)
        except KeyError:
            return False
        else:
            return True

    def __getstate__(self):
        s = {
            "_arch": self._arch,
            "_backers": self._backers,
            "_pointer": self._pointer,
            "_root": self._root,
            "consecutive": self.consecutive,
            "min_addr": self.min_addr,
            "max_addr": self.max_addr,
        }

        return s

    def __setstate__(self, s):
        self._arch = s["_arch"]
        self._backers = s["_backers"]
        self._pointer = s["_pointer"]
        self._root = s["_root"]
        self.consecutive = s["consecutive"]
        self.min_addr = s["min_addr"]
        self.max_addr = s["max_addr"]

    def backers(self, addr=0):
        """
        Iterate through each backer for this clemory and all its children, yielding tuples of
        ``(start_addr, backer)`` where each backer is a bytearray.

        :param addr:    An optional starting address - all backers before and not including this
                        address will be skipped.
        """
        started = False
        for start, backer in self._backers:
            if not started:
                end = start + backer.max_addr if isinstance(backer, Clemory) else start + len(backer)
                if addr >= end:
                    continue
                started = True
            if isinstance(backer, Clemory):
                for s, b in backer.backers(addr - start):
                    yield s + start, b
            else:
                yield start, backer

    def load(self, addr, n):
        """
        Read up to `n` bytes at address `addr` in memory and return a bytes object.

        Reading will stop at the beginning of the first unallocated region found, or when
        `n` bytes have been read.
        """
        views = []

        for start, backer in self.backers(addr):
            if start > addr:
                break
            offset = addr - start
            if not views and offset + n < len(backer):
                return bytes(memoryview(backer)[offset : offset + n])
            size = len(backer) - offset
            views.append(memoryview(backer)[offset : offset + n])

            addr += size
            n -= size

            if n <= 0:
                break

        if not views:
            raise KeyError(addr)
        return b"".join(views)

    def store(self, addr, data):
        """
        Write bytes from `data` at address `addr`.

        Note: If the store runs off the end of a backer and into unbacked space, this function
        will update the backer but also raise ``KeyError``.
        """
        for start, backer in self.backers(addr):
            if start > addr:
                raise KeyError(addr)
            offset = addr - start
            size = len(backer) - offset
            backer[offset : offset + len(data)] = data if len(data) <= size else data[:size]

            addr += size
            data = data[size:]

            if not data:
                break

        if data:
            raise KeyError(addr)

    def find(self, data, search_min=None, search_max=None):
        """
        Find all occurances of a bytestring in memory.

        :param bytes data:          The bytestring to search for
        :param int search_min:      Optional: The first address to include as valid
        :param int search_max:      Optional: The last address to include as valid
        :return Iterator[int]:      Iterates over addresses at which the bytestring occurs
        """
        if search_min is None:
            search_min = self.min_addr
        if search_max is None:
            search_max = self.max_addr

        for start, backer in self._backers:
            if isinstance(backer, Clemory):
                if search_max < backer.min_addr + start or search_min > backer.max_addr + start:
                    continue
                yield from (addr + start for addr in backer.find(data, search_min - start, search_max - start))
            elif isinstance(backer, list):
                raise TypeError("find is not supported for list-backed clemories")
            else:
                if search_max < start or search_min > start + len(data):
                    continue
                ptr = search_min - start - 1
                while True:
                    ptr += 1
                    ptr = backer.find(data, max(0, ptr))
                    if ptr == -1 or ptr + len(data) > search_max - start - 1:
                        break
                    yield ptr + start

    def _update_min_max(self):
        """
        Update the three properties of Clemory: consecutive, min_addr, and max_addr.
        """

        is_consecutive = True
        next_start = None
        min_addr, max_addr = None, None

        for start, backer in self._backers:
            if min_addr is None:
                min_addr = start

            if next_start is not None:
                # Check the predicted start equals to the real one
                if next_start != start:
                    is_consecutive = False

            if isinstance(backer, bytearray | list | mmap):
                backer_length = len(backer)
                # Update max_addr
                if max_addr is None or start + backer_length > max_addr:
                    max_addr = start + backer_length
                # Update the predicted starting address
                next_start = start + backer_length

            elif isinstance(backer, Clemory):
                if backer.max_addr is not None and backer.min_addr is not None:
                    # Update max_addr
                    if max_addr is None or start + backer.max_addr > max_addr:
                        max_addr = start + backer.max_addr
                    if backer.min_addr > 0:
                        is_consecutive = False
                    # Update the predicted starting address
                    next_start = start + backer.max_addr

                if not backer.consecutive:
                    is_consecutive = False
            else:
                raise TypeError(f"Unsupported backer type {type(backer)}.")

        self.consecutive = is_consecutive
        self.min_addr = min_addr
        self.max_addr = max_addr


class ClemoryView(ClemoryBase):
    """
    A Clemory which presents a subset of another Clemory as an address space.
    """

    def __init__(self, backer, start, end, offset=0):
        """
        :param backer:  The parent clemory to use
        :param start:   The address in the parent to start at
        :param end:     The address in the parent to end at (exclusive)
        :param offset:  Where the address space should start in this Clemory. Default 0.
        """
        super().__init__(backer._arch)
        self._backer = backer
        self._start = start
        self._end = end
        self._offset = offset
        self._endoffset = offset + (end - start)
        self._rebase = self._start - self._offset

    def __getitem__(self, k):
        if not self._offset <= k < self._endoffset:
            raise KeyError(k)
        return self._backer[k + self._rebase]

    def __setitem__(self, k, v):
        if not self._offset <= k < self._endoffset:
            raise KeyError(k)
        return self._backer[k + self._rebase]

    def __contains__(self, k):
        if not self._offset <= k < self._endoffset:
            raise KeyError(k)
        return k + self._rebase in self._backer

    def backers(self, addr=0):
        for oaddr, backer in self._backer.backers(addr=addr + self._rebase):
            taddr = oaddr - self._rebase
            if self._offset <= taddr < self._endoffset and self._offset <= taddr + len(backer) - 1 < self._endoffset:
                yield taddr, backer
            elif taddr >= self._endoffset or taddr + len(backer) - 1 < self._offset:
                continue
            else:
                # clamp it via a memoryview
                view = memoryview(backer)
                if taddr + len(backer) - 1 >= self._endoffset:
                    clamp_end = len(backer) - self._endoffset + taddr
                else:
                    clamp_end = len(backer)

                if taddr < self._offset:
                    clamp_start = self._offset - taddr
                else:
                    clamp_start = 0

                yield taddr, view[clamp_start:clamp_end]

    def load(self, addr, n):
        if n == 0:
            return b""
        if not self._offset <= addr < self._endoffset:
            raise KeyError(addr)
        if not self._offset <= addr + n - 1 < self._endoffset:
            raise KeyError(addr + n - 1)
        return self._backer.load(addr + self._rebase, n)

    def store(self, addr, data):
        if not data:
            return
        if not self._offset <= addr < self._endoffset:
            raise KeyError(addr)
        if not self._offset <= addr + len(data) - 1 < self._endoffset:
            raise KeyError(addr + len(data) - 1)
        self._backer.store(addr + self._rebase, data)

    def find(self, data, search_min=None, search_max=None):
        if search_min is None or search_min < self._start:
            search_min = self._start
        if search_max is None or search_max > self._end:
            search_max = self._end
        return self._backer.find(data, search_min=search_min + self._rebase, search_max=search_max + self._rebase)


class ClemoryTranslator(ClemoryBase):
    """
    Uses a function to translate between address spaces when accessing a child clemory. Intended to be used only as
    a stream object.
    """

    def __init__(self, backer: ClemoryBase, func):
        super().__init__(backer._arch)
        self.backer = backer
        self.func = func

    def __getitem__(self, k):
        return self.backer[self.func(k)]

    def __setitem__(self, k, v):
        self.backer[self.func(k)] = v

    def __contains__(self, k):
        return self.func(k) in self.backer

    def load(self, addr, n):
        return self.backer.load(self.func(addr), n)

    def store(self, addr, data):
        self.backer.store(self.func(addr), data)

    def backers(self, addr=0):
        raise TypeError("Cannot access backers through address translation")

    def find(self, data, search_min=None, search_max=None):
        raise TypeError("Cannot perform finds through address translation")


class UninitializedClemory(Clemory):
    """
    A special kind of Clemory that acts as a placeholder for uninitialized and invalid memory.
    This is needed for the PAGEZERO segment for MachO binaries, which is 4GB worth of memory
    This does _not_ handle data being written to it, this is only for uninitialized memory that is technically occupied
    but should never be accessed
    """

    def __init__(self, arch, size):
        super().__init__(arch, root=False)
        self.max_addr = size

    def add_backer(self, start, data, overwrite=False):
        raise ValueError("Cannot add backers to an uninitialized clemory")

    def split_backer(self, addr):
        raise ValueError("This is an uninitialized clemory, it cannot be split")

    def update_backer(self, start, data):
        raise ValueError("This is an uninitialized clemory, it cannot be updated")

    def remove_backer(self, start):
        raise ValueError("This is an uninitialized clemory, backers cannot be removed")

    def backers(self, addr=0):
        """
        Technically this object has no real backer
        We could create a fake backer on demand, but that would be a waste of memory, and code like the
        function prolog discovery for MachO binaries would search 4GB worth of nullbytes for a prolog,
        which is a waste of time
        Instead we just return an empty byte array, which seems to pass the test cases
        :param addr:
        :return:
        """
        return [(0, b"")]

    def load(self, addr, n):
        return b"\x00" * n

    def store(self, addr, data):
        raise ValueError()

    def find(self, data, search_min=None, search_max=None):
        """
        The memory has no value, so matter what is searched for, it won't be found.
        :param data:
        :param search_min:
        :param search_max:
        :return:
        """
        return iter(())


class ClemoryReadOnlyView(ClemoryBase):
    """
    Represents an outermost read-only view of a Clemory object that does not allow updates. This class offers quick
    accesses to memory reads.
    """

    def __init__(self, arch, clemory: Clemory):
        super().__init__(arch)
        self._clemory = clemory
        self._flattened_backers: list[tuple[int, bytearray]] = []

        # cache
        self._last_backer_pos: int | None = None

        self._flatten_backers()

    def __getitem__(self, k) -> int:
        # check cache first
        if self._last_backer_pos is not None:
            start, data = self._flattened_backers[self._last_backer_pos]
            if 0 <= k - start < len(data):
                return data[k - start]

        idx = bisect.bisect_right(self._flattened_backers, k, key=lambda x: x[0])
        if idx > 0:
            idx -= 1
        if idx >= len(self._flattened_backers):
            raise KeyError(k)
        start, data = self._flattened_backers[idx]
        if 0 <= k - start < len(data):
            self._last_backer_pos = idx
            return data[k - start]
        raise KeyError(k)

    def __setitem__(self, k, v):
        raise NotImplementedError("ClemoryReadOnlyView does not support item assignment")

    def load(self, addr: int, n: int) -> bytes:
        """
        Read up to `n` bytes at address `addr` in memory and return a bytes object.

        Reading will stop at the beginning of the first unallocated region found, or when
        `n` bytes have been read.
        """
        # check cache first
        if self._last_backer_pos is not None:
            start, data = self._flattened_backers[self._last_backer_pos]
            if 0 <= addr - start < len(data):
                offset = addr - start
                if offset + n < len(data):
                    return bytes(memoryview(data)[offset : offset + n])

        start_pos = bisect.bisect_right(self._flattened_backers, addr, key=lambda x: x[0])
        if start_pos > 0:
            start_pos -= 1
        views = []
        for i in range(start_pos, len(self._flattened_backers)):
            start, data = self._flattened_backers[i]
            if start > addr:
                break
            offset = addr - start
            if not views and offset + n < len(data):
                # only cache if we do not need to read across backers
                self._last_backer_pos = i
                return bytes(memoryview(data)[offset : offset + n])
            size = len(data) - offset
            views.append(memoryview(data)[offset : offset + n])

            addr += size
            n -= size

            if n <= 0:
                break

        if not views:
            raise KeyError(addr)
        return b"".join(views)

    def store(self, addr, data):
        raise NotImplementedError("ClemoryReadOnlyView does not support storing")

    def backers(self, addr: int = 0):
        start_pos = bisect.bisect_right(self._flattened_backers, addr, key=lambda x: x[0])
        if start_pos > 0:
            start_pos -= 1
        for idx in range(start_pos, len(self._flattened_backers)):
            start, data = self._flattened_backers[idx]
            if start > addr:
                break
            if 0 <= addr - start < len(data):
                yield start, data

    def unpack(self, addr, fmt):
        if self._last_backer_pos is not None:
            start, data = self._flattened_backers[self._last_backer_pos]
            if 0 <= addr - start < len(data):
                try:
                    return struct.unpack_from(fmt, data, addr - start)
                except struct.error as ex:
                    if len(data) - (addr - start) >= struct.calcsize(fmt):
                        raise ex
                    raise KeyError(addr) from ex

        idx = bisect.bisect_right(self._flattened_backers, addr, key=lambda x: x[0])
        if idx > 0:
            idx -= 1
        if idx >= len(self._flattened_backers):
            raise KeyError(addr)
        start, data = self._flattened_backers[idx]
        if start > addr:
            raise KeyError(addr)
        try:
            v = struct.unpack_from(fmt, data, addr - start)
            self._last_backer_pos = idx
            return v
        except struct.error as ex:
            if len(data) - (addr - start) >= struct.calcsize(fmt):
                raise ex
            raise KeyError(addr) from ex

    def _flatten_backers(self):
        for start, backer in self._clemory.backers():
            if isinstance(backer, bytearray):
                self._flattened_backers.append((start, backer))
            elif isinstance(backer, list):
                raise TypeError("ClemoryReadOnlyView does not support list-backed clemories")
            elif isinstance(backer, Clemory):
                pass
            else:
                raise TypeError(f"Unsupported backer type {type(backer)}.")



from __future__ import annotations

import logging
import struct

from cle.backends.symbol import SymbolType

from .common import PointTo
from .simdata import SimData, register

log = logging.getLogger(name=__name__)

#
# Here, we define a specific structure (part of it at least) for the FILE structure.
# These offsets are copied from glibc for maximum compatibility, but we are effectively
# implementing SOME libc with these symbols, so we need SOME implementation of FILE.
#
# this is supposed to be an opaque structure, the internals of which are only cared about
# by an angr simprocedure or whatever implements the fread/fwrite/etc we're linking to. And since we're linking to
# this crap instead of a real stdin/stdout/etc, someone in python land will probably be the guy which needs ABI
# compatibility with us.
#
# however, it is also a desirable property that this is abi-compatible with glibc or something so the someone in python
# land could use this to interface with the "real" structure, which would be filled out by someone other than the
# below code. To this end we so far only have the fileno, but we could add more things like buffers
#

_IO_FILE = {
    "MIPS32": {
        "size": 148,
        "fd": 0x38,
    },
    "X86": {
        "size": 148,
        "fd": 0x38,
    },
    "AMD64": {
        "size": 216,
        "fd": 0x70,
    },
    # Bionic libc does not use __IO_FILE
    # Refer to http://androidxref.com/5.1.1_r6/xref/bionic/libc/include/stdio.h
    # __sFILE replaces __IO_FILE
    # _file replaces _fileno
    "ARM": {
        "size": 84,
        "fd": 0x0E,
    },
    "AARCH64": {
        "size": 152,
        "fd": 0x14,
    },
}

_IO_FILE["ARMEL"] = _IO_FILE["ARM"]
_IO_FILE["ARMHF"] = _IO_FILE["ARM"]


def io_file_data_for_arch(arch):
    if arch.name not in _IO_FILE:
        log.error("missing _IO_FILE offsets for arch: %s", arch.name)
        return _IO_FILE["AMD64"]
    return _IO_FILE[arch.name]


class IoFilePointer(PointTo):
    libname = "libc.so.6"
    pointto_type = SymbolType.TYPE_OBJECT


class IoStdinPointer(IoFilePointer):
    name = "stdin"
    pointto_name = "_io_stdin"


class IoStdoutPointer(IoFilePointer):
    name = "stdout"
    pointto_name = "_io_stdout"


class IoStderrPointer(IoFilePointer):
    name = "stderr"
    pointto_name = "_io_stderr"


class IoFile(SimData):
    libname = "libc.so.6"
    type = SymbolType.TYPE_OBJECT
    fd: int = NotImplemented

    @classmethod
    def static_size(cls, owner):
        return io_file_data_for_arch(owner.arch)["size"]

    # the canonical verision of this should be the FILEBUF_LITERAL macro from glibc
    # for maximum hyperrealism we could have a dependency on the IO_jumps table which would have dependencies on
    # all the functions we could care about which would be implemented by simprocedures
    # but that's way overkill. see above discussion.
    def value(self):
        val = bytearray(self.size)
        struct.pack_into(self.owner.arch.struct_fmt(size=4), val, io_file_data_for_arch(self.owner.arch)["fd"], self.fd)
        struct.pack_into(self.owner.arch.struct_fmt(size=4), val, 0, 0xFBAD2088)
        return bytes(val)


class IoStdin(IoFile):
    name = "_io_stdin"
    fd = 0


class IoStdout(IoFile):
    name = "_io_stdout"
    fd = 1


class IoStderr(IoFile):
    name = "_io_stderr"
    fd = 2


register(IoStdinPointer)
register(IoStdoutPointer)
register(IoStderrPointer)
register(IoStdin)
register(IoStdout)
register(IoStderr)



from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING

from cle.address_translator import AT

if TYPE_CHECKING:
    from .backend import Backend

log = logging.getLogger(name=__name__)

__all__ = [
    "SymbolType",
    "SymbolSubType",
    "Symbol",
]


class SymbolType(Enum):
    """
    ABI-agnostic symbol types
    """

    TYPE_OTHER = 0
    TYPE_NONE = 1
    TYPE_FUNCTION = 2
    TYPE_OBJECT = 3
    TYPE_SECTION = 4
    TYPE_TLS_OBJECT = 5


class SymbolSubType(Enum):
    """
    Abstract base class for ABI-specific symbol types
    """

    def to_base_type(self) -> SymbolType:  # pylint: disable=no-self-use
        """
        A subclass' ABI-specific mapping to :SymbolType:
        """
        raise ValueError("Abstract base class SymbolSubType has no base_type")


class Symbol:
    """
    Representation of a symbol from a binary file. Smart enough to rebase itself.

    There should never be more than one Symbol instance representing a single symbol. To make sure of this, only use
    the :meth:`cle.backends.Backend.get_symbol()` to create new symbols.

    :ivar owner:            The object that contains this symbol
    :vartype owner:         cle.backends.Backend
    :ivar str name:         The name of this symbol
    :ivar int addr:         The un-based address of this symbol, an RVA
    :ivar int size:         The size of this symbol
    :ivar _type:            The ABI-agnostic type of this symbol
    :ivar bool resolved:    Whether this import symbol has been resolved to a real symbol
    :ivar resolvedby:       The real symbol this import symbol has been resolve to
    :vartype resolvedby:    None or cle.backends.Symbol
    :ivar str resolvewith:  The name of the library we must use to resolve this symbol, or None if none is required.
    """

    def __init__(self, owner: Backend, name: str, relative_addr: int, size: int, sym_type: SymbolType):
        """
        Not documenting this since if you try calling it, you're wrong.
        """
        self.owner: Backend = owner
        self.name = name
        self.relative_addr = relative_addr
        self.size = size
        self._type: SymbolType = SymbolType(sym_type)
        self.resolved = False
        self.resolvedby = None

    def __repr__(self):
        if self.is_import:
            return f'<Symbol "{self.name}" in {self.owner.binary_basename} (import)>'
        else:
            return f'<Symbol "{self.name}" in {self.owner.binary_basename} at {self.rebased_addr:#x}>'

    def resolve(self, obj):
        self.resolved = True
        self.resolvedby = obj
        self.owner.resolved_imports.append(self)

    @property
    def type(self) -> SymbolType:
        """
        The ABI-agnostic SymbolType. Must be overridden by derived types.
        """
        return self._type

    @property
    def subtype(self) -> SymbolSubType:
        """
        A subclass' ABI-specific types
        """
        raise ValueError("Base class Symbol has no subtype")

    @property
    def rebased_addr(self):
        """
        The address of this symbol in the global memory space
        """
        return AT.from_rva(self.relative_addr, self.owner).to_mva()

    @property
    def linked_addr(self):
        return AT.from_rva(self.relative_addr, self.owner).to_lva()

    @property
    def is_function(self):
        """
        Whether this symbol is a function
        """
        return self.type is SymbolType.TYPE_FUNCTION

    # These may be overridden in subclasses
    is_static = False
    is_common = False
    is_import = False
    is_export = False
    is_local = False
    is_weak = False
    is_extern = False
    is_forward = False

    def resolve_forwarder(self):
        """
        If this symbol is a forwarding export, return the symbol the forwarding refers to, or None if it cannot be found
        """
        return self

    # compatibility layer

    _complained_owner = False

    @property
    def owner_obj(self):
        if not Symbol._complained_owner:
            Symbol._complained_owner = True
            log.critical("Deprecation warning: use symbol.owner instead of symbol.owner_obj")
        return self.owner

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if k != "owner"}

    def __setstate__(self, state):
        self.__dict__.update(state)



from __future__ import annotations

from cle.backends.region import Section, Segment


def maybedecode(string):
    return string if isinstance(string, str) else string.decode()


class ELFSegment(Segment):
    """
    Represents a segment for the ELF format.
    """

    def __init__(self, readelf_seg, relro=False):
        self.flags = readelf_seg.header.p_flags
        self.relro = relro
        super().__init__(
            readelf_seg.header.p_offset,
            readelf_seg.header.p_vaddr,
            readelf_seg.header.p_filesz,
            readelf_seg.header.p_memsz,
        )

    @property
    def is_readable(self):
        return self.flags & 4 != 0

    @property
    def is_writable(self):
        return self.flags & 2 != 0

    @property
    def is_executable(self):
        return self.flags & 1 != 0

    @property
    def is_relro(self):
        return self.relro


class ELFSection(Section):
    SHF_WRITE = 0x1
    SHF_ALLOC = 0x2
    SHF_EXECINSTR = 0x4
    SHF_STRINGS = 0x20
    SHT_NULL = "SHT_NULL"

    def __init__(self, readelf_sec, remap_offset=0):
        super().__init__(
            maybedecode(readelf_sec.name),
            readelf_sec.header.sh_offset,
            readelf_sec.header.sh_addr + remap_offset,
            readelf_sec.header.sh_size,
        )

        self.type = readelf_sec.header.sh_type
        self.entsize = readelf_sec.header.sh_entsize
        self.flags = readelf_sec.header.sh_flags
        self.link = readelf_sec.header.sh_link
        self.info = readelf_sec.header.sh_info
        self.align = readelf_sec.header.sh_addralign
        self.remap_offset = remap_offset

    @property
    def is_readable(self):
        return True

    @property
    def is_active(self):
        return self.type != self.SHT_NULL

    @property
    def is_writable(self):
        return self.flags & self.SHF_WRITE != 0

    @property
    def occupies_memory(self):
        return self.flags & self.SHF_ALLOC != 0 and self.memsize > 0

    @property
    def is_executable(self):
        return self.flags & self.SHF_EXECINSTR != 0

    @property
    def is_strings(self):
        return self.flags & self.SHF_STRINGS != 0

    @property
    def only_contains_uninitialized_data(self):
        return self.type == "SHT_NOBITS"
